--
-- dpmem_pkg.vhd - dual-port FPGA-internal memory component package
--
-- SYNTHESIZABLE
--

library ieee;
use ieee.std_logic_1164.all;

package dpmem is

    constant dpmem_family_unknown : integer := -1;
    constant dpmem_family_virtex : integer := 1;    -- Virtex/Virtex-E/Virtex-EM
    constant dpmem_family_virtex2 : integer := 2;   -- Virtex-II
    constant dpmem_family_virtex2p : integer := 3;  -- Virtex-IIPro
    constant dpmem_family_virtex4 : integer := 4;   -- Virtex-4
    constant dpmem_family_virtex5 : integer := 5;   -- Virtex-5

    -- returns size (in bits) of blockram
    function blockram_data_size(
        family : in integer)
    return natural;

    function blockram_is_valid_width(
        family : in integer;
        width  : in natural)
    return boolean;

    -- returns maximum depth (in words) of blockram
    function blockram_max_depth(
        family : in integer)
    return natural;

    -- returns minimum depth (in words) of blockram
    function blockram_min_depth(
        family : in integer)
    return natural;

    -- returns maximum data width (in bits) of blockram
    function blockram_max_data_width(
        family : in integer)
    return natural;

    -- returns minimum data width (in bits) of blockram
    function blockram_min_data_width(
        family : in integer)
    return natural;

    -- returns ratio of data bits to parity bits in blockram
    function blockram_parity_ratio(
        family : in integer)
    return natural;

    component dpbram
        generic(
            order0      : in  natural;
            width0      : in  natural;
            order1      : in  natural;
            width1      : in  natural;
            -- synopsys translate_off
            init        : in  std_logic_vector;
            -- synopsys translate_on
            family      : in  integer;
            depth_waste : in  boolean);
        port(
            rst : in  std_logic;
            k0  : in  std_logic;
            sr0 : in  std_logic;
            en0 : in  std_logic;
            w0  : in  std_logic;
            a0  : in  std_logic_vector(order0 - 1 downto 0);
            d0  : in  std_logic_vector(width0 - 1 downto 0);
            q0  : out std_logic_vector(width0 - 1 downto 0);
            k1  : in  std_logic;
            sr1 : in  std_logic;
            en1 : in  std_logic;
            w1  : in  std_logic;
            a1  : in  std_logic_vector(order1 - 1 downto 0);
            d1  : in  std_logic_vector(width1 - 1 downto 0);
            q1  : out std_logic_vector(width1 - 1 downto 0));
    end component;

    component dpbram0
        generic(
            order0      : in  natural;
            width0      : in  natural;
            order1      : in  natural;
            width1      : in  natural;
            -- synopsys translate_off
            init        : in  std_logic_vector;
            -- synopsys translate_on
            family      : in  integer);
        port(
            rst : in  std_logic;
            k0  : in  std_logic;
            sr0 : in  std_logic;
            en0 : in  std_logic;
            w0  : in  std_logic;
            a0  : in  std_logic_vector(order0 - 1 downto 0);
            d0  : in  std_logic_vector(width0 - 1 downto 0);
            q0  : out std_logic_vector(width0 - 1 downto 0);
            k1  : in  std_logic;
            sr1 : in  std_logic;
            en1 : in  std_logic;
            w1  : in  std_logic;
            a1  : in  std_logic_vector(order1 - 1 downto 0);
            d1  : in  std_logic_vector(width1 - 1 downto 0);
            q1  : out std_logic_vector(width1 - 1 downto 0));
    end component;

    component distmem
        generic(
            order       : in  natural;          -- number of address bits
            width       : in  natural;          -- width of port
            -- synopsys translate_off
            init        : in  std_logic_vector; -- initial contents of RAM
            -- synopsys translate_on
            family      : in  integer);         -- FPGA family to target
        port(
            rst : in  std_logic;
            k0  : in  std_logic;
            sr0 : in  std_logic;
            en0 : in  std_logic;
            w0  : in  std_logic;
            a0  : in  std_logic_vector(order - 1 downto 0);
            d0  : in  std_logic_vector(width - 1 downto 0);
            q0  : out std_logic_vector(width - 1 downto 0);
            k1  : in  std_logic;
            sr1 : in  std_logic;
            en1 : in  std_logic;
            a1  : in  std_logic_vector(order - 1 downto 0);
            q1  : out std_logic_vector(width - 1 downto 0));
    end component;

    component distmem0
        generic(
            order       : in  natural;          -- number of address bits
            width       : in  natural;          -- width of port
            -- synopsys translate_off
            init        : in  std_logic_vector; -- initial contents of RAM
            -- synopsys translate_on
            family      : in  integer);         -- FPGA family to target
        port(
            rst : in  std_logic;
            k0  : in  std_logic;
            sr0 : in  std_logic;
            en0 : in  std_logic;
            w0  : in  std_logic;
            a0  : in  std_logic_vector(order - 1 downto 0);
            d0  : in  std_logic_vector(width - 1 downto 0);
            q0  : out std_logic_vector(width - 1 downto 0);
            k1  : in  std_logic;
            sr1 : in  std_logic;
            en1 : in  std_logic;
            a1  : in  std_logic_vector(order - 1 downto 0);
            q1  : out std_logic_vector(width - 1 downto 0));
    end component;

end;

package body dpmem is

    function blockram_data_size(
        family : in integer)
    return natural is begin
        case family is
            when dpmem_family_virtex2 | dpmem_family_virtex2p | dpmem_family_virtex4 =>
                return 16384;
            when dpmem_family_virtex =>
                return 4096;
            when others =>
                return 0;
        end case;
    end;

    function blockram_is_valid_width(
        family : in integer;
        width  : in natural)
    return boolean is begin
        case family is
            when dpmem_family_virtex2 | dpmem_family_virtex2p | dpmem_family_virtex4 =>
                case width is
                    when 1 =>
                        return true;
                    when 2 =>
                        return true;
                    when 4 =>
                        return true;
                    when 9 =>
                        return true;
                    when 18 =>
                        return true;
                    when 36 =>
                        return true;
                    when others =>
                        return false;
                end case;
            when dpmem_family_virtex =>
                case width is
                    when 1 =>
                        return true;
                    when 2 =>
                        return true;
                    when 4 =>
                        return true;
                    when 8 =>
                        return true;
                    when 16 =>
                        return true;
                    when others =>
                        return false;
                end case;
            when others =>
                return false;
        end case;
    end;

    function blockram_max_depth(
        family : in integer)
    return natural is begin
        case family is
            when dpmem_family_virtex2 | dpmem_family_virtex2p | dpmem_family_virtex4 =>
                return 16384;
            when dpmem_family_virtex =>
                return 4096;
            when others =>
                assert false
                    report "blockram_max_depth: invalid family"
                    severity failure;
                return 0;
        end case;
    end;

    function blockram_min_depth(
        family : in integer)
    return natural is begin
        case family is
            when dpmem_family_virtex2 | dpmem_family_virtex2p | dpmem_family_virtex4 =>
                return 512;
            when dpmem_family_virtex =>
                return 256;
            when others =>
                assert false
                    report "blockram_min_depth: invalid family"
                    severity failure;
                return 0;
        end case;
    end;

    function blockram_max_data_width(
        family : in integer)
    return natural is begin
        case family is
            when dpmem_family_virtex2 | dpmem_family_virtex2p | dpmem_family_virtex4 =>
                return 32;
            when dpmem_family_virtex =>
                return 16;
            when others =>
                assert false
                    report "blockram_max_data_width: invalid family"
                    severity failure;
                return 0;
        end case;
    end;

    function blockram_min_data_width(
        family : in integer)
    return natural is begin
        case family is
            when dpmem_family_virtex2 | dpmem_family_virtex2p | dpmem_family_virtex4 =>
                return 1;
            when dpmem_family_virtex =>
                return 1;
            when others =>
                assert false
                    report "blockram_min_data_width: invalid family"
                    severity failure;
                return 0;
        end case;
    end;

    function blockram_parity_ratio(
        family : in integer)
    return natural is begin
        case family is
            when dpmem_family_virtex2 | dpmem_family_virtex2p | dpmem_family_virtex4 =>
                return 8;
            when dpmem_family_virtex =>
                return 65536;
            when others =>
                assert false
                    report "blockram_parity_ratio: invalid family"
                    severity failure;
                return 0;
        end case;
    end;

end;
