--
-- oserdes_dqs.vhd - ODDR emulation of OSERDES instance for Virtex-4 / Virtex-5 DDR-II SDRAM interface
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

-- synopsys translate_off
library unisim;
use unisim.vcomponents.all;
library std;
use std.textio.all;
-- synopsys translate_on

library work;
use work.memif.all;

entity oserdes_dqs is
    port (
        clk        : in  std_logic;
        clk2x      : in  std_logic;
        sr         : in  std_logic;
        ce         : in  std_logic;
        d          : in  std_logic;
        db         : in  std_logic;
        t          : in  std_logic;
        q          : out std_logic;
        tq         : out std_logic;
        qb         : out std_logic;
        tqb        : out std_logic);
end entity;

architecture mixed of oserdes_dqs is

    signal d_q, db_q : std_logic;
    signal t_q0, t_q1 : std_logic;

    signal logic0 : std_logic;
    signal logic1 : std_logic;
    signal logicx : std_logic;

    component ODDR
        generic(
            DDR_CLK_EDGE : in  string := "OPPOSITE_EDGE";
            INIT         : in  bit := '1';
            SRTYPE       : in  string := "SYNC");
        port(
            D1           : in  std_logic;
            D2           : in  std_logic;
            CE           : in  std_logic;
            C            : in  std_logic;
            S            : in  std_logic;
            R            : in  std_logic;
            Q            : out std_logic);
    end component;

    --
    -- XST synthesis constraints, required to prevent optimization of
    -- flip-flops that would otherwise prevent IOB packing.
    --
    attribute equivalent_register_removal : string;
    attribute equivalent_register_removal of t_q0 : signal is "no";
    attribute equivalent_register_removal of t_q1 : signal is "no";
    attribute equivalent_register_removal of tq : signal is "no";
    attribute equivalent_register_removal of d_q : signal is "no";
    attribute equivalent_register_removal of db_q : signal is "no";

    attribute DDR_CLK_EDGE : string;
    attribute DDR_CLK_EDGE of U0 : label is "SAME_EDGE";
    attribute DDR_CLK_EDGE of U1 : label is "SAME_EDGE";
    attribute DDR_CLK_EDGE of U2 : label is "SAME_EDGE";
    attribute DDR_CLK_EDGE of U3 : label is "SAME_EDGE";
    attribute INIT : bit;
    attribute INIT of U0 : label is '1';
    attribute INIT of U1 : label is '1';
    attribute INIT of U2 : label is '0';
    attribute INIT of U3 : label is '1';
    attribute SRTYPE : string;
    attribute SRTYPE of U0 : label is "ASYNC";
    attribute SRTYPE of U1 : label is "ASYNC";
    attribute SRTYPE of U2 : label is "ASYNC";
    attribute SRTYPE of U3 : label is "ASYNC";

begin

    logic0 <= '0';
    logic1 <= '1';
    logicx <= '-';

    delay_t : process(clk)
    begin
        if clk'event and clk = '0' then
            if sr = '1' then
                t_q0 <= '1';
                t_q1 <= '1';
            else
                t_q0 <= t;
                t_q1 <= t_q0;
            end if;
        end if;
    end process;

    delay_d : process(clk)
    begin
        if clk'event and clk = '1' then
            d_q <= d;
            db_q <= db;
        end if;
    end process;

    U0 : ODDR
        generic map(
            DDR_CLK_EDGE => "SAME_EDGE",
            INIT => '1',
            SRTYPE => "ASYNC")
        port map(
            D1 => t_q1,
            D2 => t_q1,
            CE => logic1,
            C => clk2x,
            S => logic0,
            R => logic0,
            Q => tq);

    U1 : ODDR
        generic map(
            DDR_CLK_EDGE => "SAME_EDGE",
            INIT => '1',
            SRTYPE => "ASYNC")
        port map(
            D1 => t_q1,
            D2 => t_q1,
            CE => logic1,
            C => clk2x,
            S => logic0,
            R => logic0,
            Q => tqb);

    U2 : ODDR
        generic map(
            DDR_CLK_EDGE => "SAME_EDGE",
            INIT => '0',
            SRTYPE => "ASYNC")
        port map(
            D1 => logic0,
            D2 => d_q,
            CE => logic1,
            C => clk2x,
            S => logic0,
            R => logic0,
            Q => q);

    U3 : ODDR
        generic map(
            DDR_CLK_EDGE => "SAME_EDGE",
            INIT => '1',
            SRTYPE => "ASYNC")
        port map(
            D1 => logic1,
            D2 => db_q,
            CE => logic1,
            C => clk2x,
            S => logic0,
            R => logic0,
            Q => qb);

end mixed;
