--
-- ddr2sram_port.vhd - DDR-II SRAM interface
--
-- SYNTHESIZABLE
--
-- The following inputs control the mode of this memory interface. Except where
-- noted, they do not have to be driven by constant values:
--
--   burst_len     0 => Burst length 2
--                 1 => Burst length 4 (also works with BL2 devices)
--
--   dll_off       0 => DLL enabled (normal operation) 
--                 1 => DLL disabled
--                 (Note: the normal setting is '0', to enable the DLL. Disabling
--                 the DLL may be necessary when operating the device at a clock
--                 frequency below its minimum frequency).
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

-- synopsys translate_off
library std;
use std.textio.all;
-- synopsys translate_on

-- synopsys translate_off
library unisim;
use unisim.vcomponents.all;
-- synopsys translate_on

library work;
use work.memif.all;

entity ddr2sram_port is
    generic(
        family          : in    integer := memif_family_auto;
        model           : in    integer;
        ra_width        : in    natural;
        rc_width        : in    natural;
        rd_width        : in    natural;
        a_width         : in    natural;
        d_width         : in    natural;
        tag_width       : in    natural);
    port(
        clk0            : in    std_logic;                                   -- Clock
        clk90           : in    std_logic;
        clk180          : in    std_logic;
        clk270          : in    std_logic;
        rst             : in    std_logic;                                   -- Aynchronous reset
        sr              : in    std_logic;                                   -- Synchronous reset
        ce              : in    std_logic;                                   -- Command enable, qualified by 'ready'
        tag             : in    std_logic_vector(tag_width - 1 downto 0);    -- Tag, qualified by 'ce and not w')
        w               : in    std_logic;                                   -- 1 => write, qualified by 'ce')
        a               : in    std_logic_vector(a_width - 1 downto 0);      -- Address, qualified by 'ce'
        d               : in    std_logic_vector(d_width - 1 downto 0);      -- Write data, qualified by 'ce and w'
        be              : in    std_logic_vector(d_width / 8 - 1 downto 0);  -- Write byte enables, qualified by 'ce and w'
        q               : out   std_logic_vector(d_width - 1 downto 0);      -- Read data, qualified by 'valid'
        qtag            : out   std_logic_vector(tag_width - 1 downto 0);    -- Read tag, qualified by 'valid'
        valid           : out   std_logic;                                   -- 1 => 'q' and 'qtag' are valid
        ready           : out   std_logic;                                   -- When low, indicates that 'ce' is ignored
        -- Sideband control signals, must be stable during operation
        burst_len       : in    std_logic;                                   -- Supported burst length: 0 => 2 only, 1 => 2 & 4
        dll_off         : in    std_logic;                                   -- Disables DLL in SSRAM device when high
        -- To/from memories
        ra              : out   std_logic_vector(ra_width - 1 downto 0);
        rc              : inout std_logic_vector(rc_width - 1 downto 0);
        rd              : inout std_logic_vector(rd_width - 1 downto 0));
end entity;

architecture mixed of ddr2sram_port is

    constant this_module : string := "ddr2sram_port";

    -- synopsys translate_off
    function conv_string(
        val : in integer)
    return string is
        variable l : line;
        variable len : integer;
        variable s : string(1 to 64);
    begin
        write(l, val);
        len := l'length;
        read(l, s(1 to len));
        return s(1 to len);
    end;
    -- synopsys translate_on

    -- Get model-dependent constants
    constant use_inverted_clocks : boolean := memif_use_inverted_clocks(family, model);
    constant capture_180deg : boolean := memif_ddr2sram_capture_180deg(model);

    constant be_width : natural := d_width / 8;
    constant cmd_width : natural := d_width + be_width + a_width + 1 + tag_width;

    signal k0 : std_logic;
    signal k90 : std_logic;
    signal k180 : std_logic;
    signal k270 : std_logic;

    -- Signals to/from command FIFO (i.e. stage 0 of command pipeline)
    signal cmd_d : std_logic_vector(cmd_width - 1 downto 0);
    signal cmd_q : std_logic_vector(cmd_width - 1 downto 0);
    signal cmd_q_d : std_logic_vector(d_width - 1 downto 0);
    signal cmd_q_be : std_logic_vector(d_width / 8 - 1 downto 0);
    signal cmd_q_a : std_logic_vector(ra_width - 2 downto 0);
    signal cmd_q_w : std_logic;
    signal cmd_q_tag : std_logic_vector(tag_width - 1 downto 0);
    signal cmd_empty : std_logic;
    signal cmd_read : std_logic;

    signal cmd_q1_a : std_logic_vector(ra_width - 2 downto 0);
    signal cmd_q1_d : std_logic_vector(d_width - 1 downto 0);
    signal cmd_q2_d : std_logic_vector(d_width - 1 downto 0);
    signal cmd_q3_d : std_logic_vector(d_width - 1 downto 0);
    signal cmd_q4_d : std_logic_vector(d_width - 1 downto 0);
    signal cmd_q1_be_l : std_logic_vector(d_width / 8 - 1 downto 0);
    signal cmd_q2_be_l : std_logic_vector(d_width / 8 - 1 downto 0);
    signal cmd_q3_be_l : std_logic_vector(d_width / 8 - 1 downto 0);
    signal cmd_q4_be_l : std_logic_vector(d_width / 8 - 1 downto 0);
    signal cmd_q1_w : std_logic;
    signal cmd_q1_tag : std_logic_vector(tag_width - 1 downto 0);
    signal cmd_q2_tag : std_logic_vector(tag_width - 1 downto 0);

    -- Burst control
    signal b_idle : std_logic;
    signal full : std_logic;
    signal ce_b_hit : std_logic;
    signal b_hit : std_logic;
    signal b_write : std_logic;
    signal b_addr : std_logic_vector(ra_width - 1 downto 1);
    signal b_next_addr : std_logic_vector(ra_width - 1 downto 1);
    signal do_ld_l : std_logic;
    signal do_w_l : std_logic;
    signal do_wbe : std_logic;
    signal do_read : std_logic;
    signal do_write : std_logic;
    signal do_write_q : std_logic;

    type tag_pipe_t is array(0 to 7) of std_logic_vector(tag_width - 1 downto 0);
    signal tag_pipe : tag_pipe_t;

    signal valid_pipe : std_logic_vector(7 downto 0);

    signal i_q : std_logic_vector(d_width - 1 downto 0);
    signal i_qtag : std_logic_vector(tag_width - 1 downto 0);
    signal i_valid : std_logic;

    signal ld_o_l : std_logic;
    signal w_o_l : std_logic;
    signal k_o : std_logic;
    signal k_o_l : std_logic;
    signal c_o : std_logic;
    signal c_o_l : std_logic;
    signal doff_o_l : std_logic;
    signal be_o_l : std_logic_vector(rd_width / 8 - 1 downto 0);
    signal rd_i0, rd_i0_q, rd_i1, rd_i1_q : std_logic_vector(rd_width - 1 downto 0);

    signal ra_t : std_logic_vector(ra_width - 1 downto 0);
    signal ra_o : std_logic_vector(ra_width - 1 downto 0);
    signal rc_t : std_logic_vector(rc_width - 1 downto 0);
    signal rc_o : std_logic_vector(rc_width - 1 downto 0);
    signal rc_i : std_logic_vector(rc_width - 1 downto 0);
    signal rd_t : std_logic_vector(rd_width - 1 downto 0);
    signal rd_o : std_logic_vector(rd_width - 1 downto 0);
    signal rd_i : std_logic_vector(rd_width - 1 downto 0);

    signal logic0, logic1 : std_logic;
    signal zeroes : std_logic_vector(31 downto 0);

    --
    -- XST synthesis constraints, required to prevent optimization of
    -- flip-flops that would otherwise prevent IOB packing.
    --
    attribute equivalent_register_removal : string;
    attribute equivalent_register_removal of ra_o : signal is "no";
    attribute equivalent_register_removal of rc_o : signal is "no";
    attribute equivalent_register_removal of rd_o : signal is "no";
    attribute equivalent_register_removal of rd_i : signal is "no";
    attribute equivalent_register_removal of rd_t : signal is "no";
    attribute equivalent_register_removal of k_o : signal is "no";
    attribute equivalent_register_removal of k_o_l : signal is "no";
    attribute equivalent_register_removal of c_o : signal is "no";
    attribute equivalent_register_removal of c_o_l : signal is "no";
    attribute iob : string;
    attribute iob of rd_t : signal is "false";
    attribute iob of rd_o : signal is "true";
    attribute iob of rd_i : signal is "true";
    attribute iob of be_o_l : signal is "true";
    attribute iob of w_o_l : signal is "true";
    attribute iob of ld_o_l : signal is "true";
    attribute iob of c_o : signal is "true";
    attribute iob of c_o_l : signal is "true";
    attribute iob of k_o : signal is "true";
    attribute iob of k_o_l : signal is "true";
    attribute iob of doff_o_l : signal is "true";

    component cmd_fifo
        generic(
            width         : in  natural);
        port(
            clk           : in  std_logic;
            rst           : in  std_logic;
            sr            : in  std_logic;
            w             : in  std_logic;
            d             : in  std_logic_vector(width - 1 downto 0);
            r             : in  std_logic;
            q             : out std_logic_vector(width - 1 downto 0);
            ready         : out std_logic;
            empty         : out std_logic);
    end component;

    component FDDRRSE
        -- synthesis translate_off
        generic(
            INIT : bit := '1');
        -- synthesis translate_on
        port(
            Q  : out   std_logic;
            C0 : in    std_logic;
            C1 : in    std_logic;
            CE : in    std_logic;
            D0 : in    std_logic;
            D1 : in    std_logic;
            R  : in    std_logic;
            S  : in    std_logic);
    end component;

    component OBUF
        port(
            I  : in    std_logic;
            O  : out   std_logic);
    end component;

    component IOBUF
        port(
            I  : in    std_logic;
            IO : inout std_logic;
            O  : out   std_logic;
            T  : in    std_logic);
    end component;

begin

    -- synopsys translate_off
    assert model = memif_model_admxp or model = memif_model_adpdrc2 or model = memif_model_adpxpi or
           model = memif_model_admxrc5t1
        report "*** " & this_module & ": model " & conv_string(model) & " is not supported"
        severity failure;
    -- synopsys translate_on

    logic0 <= '0';
    logic1 <= '1';
    zeroes <= (others => '0');

    q <= i_q;
    qtag <= i_qtag;
    valid <= i_valid;

    gen_ik : if not use_inverted_clocks generate
        k0 <= clk0;
        k90 <= clk90;
        k180 <= clk180;
        k270 <= clk270;
    end generate;

    gen_ik_inv : if use_inverted_clocks generate
        k0 <= clk0;
        k90 <= clk90;
        k180 <= not clk0;
        k270 <= not clk90;
    end generate;

    --
    -- Buffer incoming commands in command FIFO
    --

    cmd_d <= tag & w & a & be & d;
    cmd_read <= ce_b_hit;

    cmd_fifo_0 : cmd_fifo
        generic map(
            width => cmd_width)
        port map(
            clk => k0,
            rst => rst,
            sr => sr,
            w => ce,
            d => cmd_d,
            ready => ready,
            r => cmd_read,
            empty => cmd_empty,
            q => cmd_q);

    cmd_q_d <= cmd_q(d_width - 1 downto 0);
    cmd_q_be <= cmd_q(d_width + be_width - 1 downto d_width);
    cmd_q_a <= EXT(cmd_q(d_width + be_width + a_width - 1 downto d_width + be_width), ra_width - 1);
    cmd_q_w <= cmd_q(d_width + be_width + a_width );
    cmd_q_tag <= cmd_q(d_width + be_width + a_width + tag_width downto d_width + be_width + a_width + 1);

    --
    -- DDR SRAM burst state machine
    --

    burst_state_machine : process(rst, k0)
    begin
        if rst = '1' then
            b_idle <= '1';
            b_write <= '-';
            b_addr <= (others => '0');
        elsif k0'event and k0 = '1' then
            case b_idle is
                when '1' => -- Idle
                    if full = '1' and (cmd_q1_w = '0' or do_read = '0') then
                        b_idle <= '0';
                        b_write <= cmd_q1_w;
                        b_addr <= cmd_q1_a;
                    else
                        b_idle <= '1';
                    end if;

                when '0' => -- Bursting
                    b_idle <= '1';
                    b_addr(1) <= not b_addr(1);

                when others =>
                    b_idle <= '-';
            end case;
        end if;
    end process;

    --
    -- Full state machine
    --

    full_state_machine : process(rst, k0)
    begin
        if rst = '1' then
            full <= '0';
        elsif k0'event and k0 = '1' then
            case full is
                when '0' =>
                    if cmd_empty = '0' then
                        full <= '1';
                    else
                        full <= '0';
                    end if;

                when '1' =>
                    if cmd_empty = '1' and ((b_idle = '1' and (cmd_q1_w = '0' or do_read = '0')) or b_hit = '1') then
                        full <= '0';
                    else
                        full <= '1';
                    end if;

                when others =>
                    full <= '-';
            end case;
        end if;
    end process;

    --
    -- Burst and pipeline control logic
    --

    ce_b_hit <= (not cmd_empty and not full) or (not cmd_empty and full and ((b_idle and (not cmd_q1_w or not do_read)) or b_hit));

    b_next_addr(ra_width - 1 downto 2) <= cmd_q1_a(ra_width - 2 downto 1);
    b_next_addr(1) <= not cmd_q1_a(0);

    check_addr : process(rst, k0)
    begin
        if rst = '1' then
            cmd_q1_a <= (others => '-');
            cmd_q1_d <= (others => '-');
            cmd_q1_be_l <= (others => '1');
            cmd_q1_w <= '-';
            cmd_q1_tag <= (others => '-');
            cmd_q2_tag <= (others => '-');
            b_hit <= '-';
        elsif k0'event and k0 = '1' then
            if ce_b_hit = '1' then
                cmd_q1_a <= cmd_q_a;
                cmd_q1_d <= cmd_q_d;
                cmd_q1_be_l <= not cmd_q_be;
                cmd_q1_w <= cmd_q_w;
                cmd_q1_tag <= cmd_q_tag;
                b_hit <= full and AND_reduce(b_next_addr xnor cmd_q_a) and (cmd_q1_w xnor cmd_q_w);
            end if;

            cmd_q2_tag <= cmd_q1_tag;
        end if;
    end process;

    --
    -- Figure out if we can do a read or a write
    --

    process(rst, k0)
        variable x : std_logic;
    begin
        if rst = '1' then
            do_ld_l <= '1';
            do_w_l <= '1';
            do_wbe <= '0';
            do_read <= '0';
            do_write <= '0';
            do_write_q <= '0';
        elsif k0'event and k0 = '1' then
            x := (b_idle and full and (not cmd_q1_w or not do_read)) or not b_idle;
            do_ld_l <= not x;
            x := (b_idle and full and cmd_q1_w and not do_read) or (not b_idle and b_write);
            do_w_l <= not x;
            do_write <= x;
            do_write_q <= do_write;
            do_wbe <= full and ((b_idle and cmd_q1_w and not do_read) or (not b_idle and b_write and b_hit));
            do_read <= full and ((b_idle and not cmd_q1_w) or (not b_idle and not b_write and b_hit));
        end if;
    end process;

    --
    -- Generate 'ld_o_l' (LD#)
    --
   
    gen_ld_o_l : process(rst, k180)
    begin
        if rst = '1' then
            ld_o_l <= '1';
        elsif k180'event and k180 = '1' then
            ld_o_l <= do_ld_l;
        end if;
    end process;

    --
    -- Generate 'w_o_l' (W#)
    --
   
    gen_w_o_l : process(rst, k270)
    begin
        if rst = '1' then
            w_o_l <= '1';
        elsif k270'event and k270 = '1' then
            w_o_l <= do_w_l;
        end if;
    end process;

    --
    -- Delay byte enables and data for two cycles, since DDR-II SSRAM
    -- devices want write data two cycles after LD#
    --

    delay_data : process(k0)
    begin
        if k0'event and k0 = '1' then
            cmd_q2_d <= cmd_q1_d;
            cmd_q3_d <= cmd_q2_d;
            cmd_q4_d <= cmd_q3_d;
        end if;
    end process;

    delay_be : process(rst, k0)
    begin
        if rst = '1' then
            cmd_q2_be_l <= (others => '1');
            cmd_q3_be_l <= (others => '1');
            cmd_q4_be_l <= (others => '1');
        elsif k0'event and k0 = '1' then
            cmd_q2_be_l <= cmd_q1_be_l;
            if do_wbe = '1' then
                cmd_q3_be_l <= cmd_q2_be_l;
            else
                cmd_q3_be_l <= (others => '1');
            end if;
            cmd_q4_be_l <= cmd_q3_be_l;
        end if;
    end process;

    --
    -- Generate 'be_o_l' (BE#)
    --

    gen_be_o_l : for i in 0 to rd_width / 8 - 1 generate
        U0 : FDDRRSE
            -- synthesis translate_off
            generic map(
                INIT => '1')
            -- synthesis translate_on
            port map(
                Q => be_o_l(i),
                C0 => k0,
                C1 => k180,
                CE => logic1,
                D0 => cmd_q3_be_l(i),
                D1 => cmd_q4_be_l(i + rd_width / 8),
                R => logic0,
                S => logic0);
    end generate;

    --
    -- Generate 'rd_o' (DQ to SSRAM)
    --

    gen_rd_o : for i in 0 to rd_width - 1 generate
        U0 : FDDRRSE
            -- synthesis translate_off
            generic map(
                INIT => '1')
            -- synthesis translate_on
            port map(
                Q => rd_o(i),
                C0 => k0,
                C1 => k180,
                CE => logic1,
                D0 => cmd_q3_d(i),
                D1 => cmd_q4_d(i + rd_width),
                R => logic0,
                S => logic0);
    end generate;

    --
    -- Generate 'rd_t' (tristate control for DQ)
    --

    gen_rd_t : process(rst, k0)
    begin
        if rst = '1' then
            rd_t <= (others => '1');
        elsif k0'event and k0 = '1' then
            rd_t <= (others => not (do_write or do_write_q));
        end if;
    end process;

    --
    -- Capture logic assuming that data should be captured using 'k180' (
    -- equivalent to falling edge of 'k0')
    --

    capture_180 : if capture_180deg generate
        gen_rd_i0 : process(rst, k180)
        begin
            if rst = '1' then
                rd_i0 <= (others => '1');
            elsif k180'event and k180 = '1' then
                rd_i0 <= rd_i;
            end if;
        end process;

        gen_rd_i1 : process(rst, k0)
        begin
            if rst = '1' then
                rd_i1 <= (others => '1');
            elsif k0'event and k0 = '1' then
                rd_i1 <= rd_i;
            end if;
        end process;

        gen_rd_i0_q : process(rst, k0)
        begin
            if rst = '1' then
                rd_i0_q <= (others => '0');
            elsif k0'event and k0 = '1' then
                rd_i0_q <= rd_i0;
            end if;
        end process;

        gen_rd_i1_q : process(rst, k0)
        begin
            if rst = '1' then
                rd_i1_q <= (others => '0');
            elsif k0'event and k0 = '1' then
                rd_i1_q <= rd_i1;
            end if;
        end process;

        i_q <= EXT(rd_i1 & rd_i0_q, d_width);
    end generate;

    --
    -- Capture logic assuming that data should be captured using 'k90'
    --

    capture_90 : if not capture_180deg generate
        gen_rd_i0 : process(rst, k90)
        begin
            if rst = '1' then
                rd_i0 <= (others => '1');
            elsif k90'event and k90 = '1' then
                rd_i0 <= rd_i;
            end if;
        end process;

        gen_rd_i1 : process(rst, k270)
        begin
            if rst = '1' then
                rd_i1 <= (others => '1');
            elsif k270'event and k270 = '1' then
                rd_i1 <= rd_i;
            end if;
        end process;

        gen_rd_i0_q : process(rst, k0)
        begin
            if rst = '1' then
                rd_i0_q <= (others => '0');
            elsif k0'event and k0 = '1' then
                rd_i0_q <= rd_i0;
            end if;
        end process;

        gen_rd_i1_q : process(rst, k0)
        begin
            if rst = '1' then
                rd_i1_q <= (others => '0');
            elsif k0'event and k0 = '1' then
                rd_i1_q <= rd_i1;
            end if;
        end process;

        i_q <= EXT(rd_i1_q & rd_i0_q, d_width);
    end generate;

    --
    -- 'valid' and 'qtag' pipelines
    --

    gen_valid : process(rst, k0)
    begin
        if rst = '1' then
            valid_pipe <= (others => '0');
            i_valid <= '0';
        elsif k0'event and k0 = '1' then
            valid_pipe(0) <= do_read;
            valid_pipe(7 downto 1) <= valid_pipe(6 downto 0);
            i_valid <= valid_pipe(2);
        end if;
    end process;

    gen_qtag : process(k0)
    begin
        if k0'event and k0 = '1' then
            tag_pipe(0) <= cmd_q2_tag;
            tag_pipe(1 to 7) <= tag_pipe(0 to 6);
            i_qtag <= tag_pipe(2);
        end if;
    end process;

    --
    -- Generate clocks to SSRAM device
    --

    gen_k_o : FDDRRSE
        port map(
            Q => k_o,
            C0 => k90,
            C1 => k270,
            CE => logic1,
            D0 => logic1,
            D1 => logic0,
            R => logic0,
            S => logic0);

    gen_k_o_l : FDDRRSE
        port map(
            Q => k_o_l,
            C0 => k90,
            C1 => k270,
            CE => logic1,
            D0 => logic0,
            D1 => logic1,
            R => logic0,
            S => logic0);

    gen_c_o : FDDRRSE
        port map(
            Q => c_o,
            C0 => k90,
            C1 => k270,
            CE => logic1,
            D0 => logic1,
            D1 => logic0,
            R => logic0,
            S => logic0);

    gen_c_o_l : FDDRRSE
        port map(
            Q => c_o_l,
            C0 => k90,
            C1 => k270,
            CE => logic1,
            D0 => logic0,
            D1 => logic1,
            R => logic0,
            S => logic0);

    --
    -- Generate 'ra_o' (address to SSRAM device)
    --

    gen_ra_o : process(k180)
    begin
        if k180'event and k180 = '1' then
            ra_o(ra_width - 1 downto 1) <= b_addr;
            ra_o(0) <= '0';
        end if;
    end process;

    --
    -- Address is always driven to SSRAM device
    --

    ra_t <= (others => '0');

    --
    -- Generate 'doff_o_l' (DOFF#)
    --
    gen_doff_o_l : process(rst, k0)
    begin
        if rst = '1' then
            doff_o_l <= '1';
        elsif k0'event and k0 = '1' then
            doff_o_l <= not dll_off;
        end if;
    end process;

    --
    -- Package up control signals into 'rc_o' bus, in a model-dependent way
    --

    gen_rc_o_xp_xpi : if model = memif_model_admxp or model = memif_model_adpxpi generate
        rc_o(rd_width / 8 - 1 downto 0) <= be_o_l; -- BWE#[]
        rc_o(rd_width / 8 + 0) <= ld_o_l;   -- LD#
        rc_o(rd_width / 8 + 1) <= w_o_l;    -- R/W#
        rc_o(rd_width / 8 + 2) <= k_o;      -- K
        rc_o(rd_width / 8 + 3) <= k_o_l;    -- K#
        rc_o(rd_width / 8 + 4) <= c_o;      -- C
        rc_o(rd_width / 8 + 5) <= c_o_l;    -- C#
        rc_o(rd_width / 8 + 6) <= doff_o_l; -- DOFF#
        rc_t <= (others => '0');
    end generate;

    gen_rc_o_drc2 : if model = memif_model_adpdrc2 generate
        rc_o(rd_width / 8 - 1 downto 0) <= be_o_l; -- BWE#[]
        rc_o(rd_width / 8 + 0) <= ld_o_l;   -- LD#
        rc_o(rd_width / 8 + 1) <= w_o_l;    -- R/W#
        rc_o(rd_width / 8 + 2) <= k_o;      -- K
        rc_o(rd_width / 8 + 3) <= k_o_l;    -- K#
        rc_o(rd_width / 8 + 4) <= c_o;      -- C
        rc_o(rd_width / 8 + 5) <= c_o_l;    -- C#
        rc_o(rd_width / 8 + 6) <= '-';      -- CQ
        rc_o(rd_width / 8 + 7) <= '-';      -- CQ#
        rc_o(rd_width / 8 + 8) <= doff_o_l; -- DOFF#
        rc_t(rd_width / 8 + 5 downto 0) <= (others => '0');
        rc_t(rd_width / 8 + 7 downto rd_width / 8 + 6) <= (others => '1');
        rc_t(rd_width / 8 + 8) <= '0';
    end generate;

    gen_rc_o_xrc5t1 : if model = memif_model_admxrc5t1 generate
        rc_o(rd_width / 8 - 1 downto 0) <= be_o_l; -- BWE#[]
        rc_o(rd_width / 8 + 0) <= ld_o_l;   -- LD#
        rc_o(rd_width / 8 + 1) <= w_o_l;    -- R/W#
        rc_o(rd_width / 8 + 2) <= k_o;      -- K
        rc_o(rd_width / 8 + 3) <= k_o_l;    -- K#
        rc_o(rd_width / 8 + 4) <= '-';      -- CQ
        rc_o(rd_width / 8 + 5) <= '-';      -- CQ#
        rc_o(rd_width / 8 + 6) <= doff_o_l; -- DOFF#
        rc_t(rd_width / 8 + 3 downto 0) <= (others => '0');
        rc_t(rd_width / 8 + 5 downto rd_width / 8 + 4) <= (others => '1');
        rc_t(rd_width / 8 + 6) <= '0';
    end generate;

    gen_ra : for i in 0 to ra_width - 1 generate
        U0 : OBUF
            port map(
                I => ra_o(i),
                O => ra(i));
    end generate;

    gen_rc : for i in 0 to rc_width - 1 generate
        U0 : IOBUF
            port map(
                I => rc_o(i),
                IO => rc(i),
                O => rc_i(i),
                T => rc_t(i));
    end generate;

    gen_rd : for i in 0 to rd_width - 1 generate
        U0 : IOBUF
            port map(
                I => rd_o(i),
                IO => rd(i),
                O => rd_i(i),
                T => rd_t(i));
    end generate;

end mixed;
