--
-- ddrsdram_data.vhd - DDR SDRAM data path
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

-- synopsys translate_off
library unisim;
use unisim.vcomponents.all;
library std;
use std.textio.all;
-- synopsys translate_on

library work;
use work.memif.all;

entity ddrsdram_data is
    generic(
        family      : in  integer;
        cas_latency : in  natural;
        flight_time : in  integer;
        rd_width    : in  natural;
        d_width     : in  natural;
        tag_width   : in  natural;
        dqs_delayed : in  boolean);
    port (
        clk0        : in  std_logic;
        clk90       : in  std_logic;
        clk180      : in  std_logic;
        clk270      : in  std_logic;
        rst         : in  std_logic;
        sr          : in  std_logic;
        regd        : in  std_logic;
        oe_l        : in  std_logic_vector(rd_width - 1 downto 0);
        wr          : in  std_logic;
        read_cmd    : in  std_logic;
        d           : in  std_logic_vector(d_width - 1 downto 0);
        tag         : in  std_logic_vector(tag_width - 1 downto 0);
        dqs_i       : in  std_Logic_vector(rd_width / 8 - 1 downto 0);
        dq_i        : in  std_Logic_vector(rd_width - 1 downto 0);
        dq_o        : out std_logic_vector(rd_width - 1 downto 0);
        dq_t        : out std_Logic_vector(rd_width - 1 downto 0);
        q           : out std_logic_vector(d_width - 1 downto 0);
        qtag        : out std_logic_vector(tag_width - 1 downto 0);
        qvalid      : out std_logic);
end;

architecture mixed of ddrsdram_data is

    -- synopsys translate_off
    function conv_string(
        val : in integer)
    return string is
        variable l : line;
        variable len : integer;
        variable s : string(1 to 64);
    begin
        write(l, val);
        len := l'length;
        read(l, s(1 to len));
        return s(1 to len);
    end;
    -- synopsys translate_on

    constant dqs_width : natural := rd_width / 8;

    constant capture_latency : natural := cas_latency * 2 + flight_time;
    constant capture_edge : natural := capture_latency mod 4;
    constant capture_delay : natural := (capture_latency + 1) / 4;

    signal dqs_i_l : std_logic_vector(rd_width / 8 - 1 downto 0);
    signal dqs_i0 : std_logic_vector(rd_width / 8 - 1 downto 0);
    signal dqs_i0_l : std_logic_vector(rd_width / 8 - 1 downto 0);

    signal d_q : std_logic_vector(rd_width * 2 - 1 downto 0);
    signal d_q_sel : std_logic_vector(rd_width * 2 - 1 downto 0);

    signal q_l0, q_h0 : std_logic_vector(rd_width - 1 downto 0);
    signal q0_l0, q0_h0 : std_logic_vector(rd_width - 1 downto 0);
    signal q0_l1, q0_h1 : std_logic_vector(rd_width - 1 downto 0);
    signal q0_l2, q0_h2 : std_logic_vector(rd_width - 1 downto 0);
    signal q0 : std_logic_vector(d_width - 1 downto 0);
    signal q1_l0, q1_h0 : std_logic_vector(rd_width - 1 downto 0);
    signal q1_l1, q1_h1 : std_logic_vector(rd_width - 1 downto 0);
    signal q1_l2, q1_h2 : std_logic_vector(rd_width - 1 downto 0);
    signal q1 : std_logic_vector(d_width - 1 downto 0);
    signal q2_l0, q2_h0 : std_logic_vector(rd_width - 1 downto 0);
    signal q2_l1, q2_h1 : std_logic_vector(rd_width - 1 downto 0);
    signal q2_l2, q2_h2 : std_logic_vector(rd_width - 1 downto 0);
    signal q2 : std_logic_vector(d_width - 1 downto 0);
    signal q3_l0, q3_h0 : std_logic_vector(rd_width - 1 downto 0);
    signal q3_l1, q3_h1 : std_logic_vector(rd_width - 1 downto 0);
    signal q3_l2, q3_h2 : std_logic_vector(rd_width - 1 downto 0);
    signal q3 : std_logic_vector(d_width - 1 downto 0);
    signal delay : std_Logic_vector(7 downto 0);
    type tag_pipe_t is array(0 to 7) of std_logic_vector(tag_width - 1 downto 0);
    signal tag_pipe : tag_pipe_t;

    signal wr_q : std_logic;
    signal wr_qq : std_logic;

    component FDDRRSE
        port(
            D0 : in  std_logic;
            D1 : in  std_logic;
            C0 : in  std_logic;
            C1 : in  std_logic;
            CE : in  std_logic;
            R  : in  std_logic;
            S  : in  std_logic;
            Q  : out std_logic);
    end component;

    component ddrcapture_reg
        generic(
            family     : in integer;
            width      : in natural);
        port(
            k0         : in  std_logic;                               -- Clock for 'q0'
            k1         : in  std_logic;                               -- Clock for 'q1'
            d          : in  std_logic_vector(width - 1 downto 0);    -- Flip-flop D input
            q0         : out std_logic_vector(width - 1 downto 0);    -- Flip-flop Q output for 'k0' clock
            q1         : out std_logic_vector(width - 1 downto 0));   -- Flip-flop Q output for 'k1' clock
    end component;

    --
    -- XST synthesis constraints, required to prevent optimization of
    -- flip-flops that would otherwise prevent IOB packing.
    --
    attribute equivalent_register_removal : string;
    attribute equivalent_register_removal of dq_o : signal is "no";
    attribute equivalent_register_removal of dq_t : signal is "no";
    attribute equivalent_register_removal of q_l0 : signal is "no";
    attribute equivalent_register_removal of q_h0 : signal is "no";
    -- Ensure that critical signals are packed into IOBs
    attribute iob : string;
    attribute iob of dq_t : signal is "false";
    attribute iob of dq_o : signal is "true";
    attribute iob of q_l0 : signal is "true";
    attribute iob of q_h0 : signal is "true";

begin

    -- synopsys translate_off
    assert cas_latency <= 8
        report "*** ddrsdram_data: illegal value " & conv_string(cas_latency) & " for cas_latency : greater than 8 half-cycles"
        severity failure;

    assert cas_latency >= 4
        report "*** ddrsdram_data: illegal value " & conv_string(cas_latency) & " for cas_latency : less than 4 half-cycles"
        severity failure;
    -- synopsys translate_on

    --
    -- Instantiate DDR registers for DQ pins
    --

    gen_d_phase90 : if not dqs_delayed generate
        --
        -- Write data path (independent of CAS latency)
        --
        -- 'd' is valid when 'wr' is asserted.
        --

        delay_data1 : process(clk0)
        begin
            if clk0'event and clk0 = '1' then
                d_q <= EXT(d, rd_width * 2);
            end if;
        end process;

        delay_data2 : process(clk270)
        begin
            if clk270'event and clk270 = '1' then
                if regd = '1' then
                    d_q_sel <= d_q;
                else
                    d_q_sel <= EXT(d, rd_width * 2);
                end if;
            end if;
        end process;

        --
        -- Here, DQ and DM are generated such that DQS falls in the middle
        -- of the valid window for DQ and DM. This is necessary when the
        -- PCB traces for DQS do not introduce significant delay.
        --

        gen_ddrff : for i in 0 to rd_width - 1 generate
            U0 : FDDRRSE
                port map(
                    D0 => d_q_sel(i),
                    D1 => d_q_sel(i + rd_width),
                    C0 => clk90,
                    C1 => clk270,
                    CE => '1',
                    R  => '0',
                    S  => '0',
                    Q  => dq_o(i));
        end generate;
    end generate;

    gen_d_phase180 : if dqs_delayed generate
        --
        -- Write data path (independent of CAS latency)
        --
        -- 'd' is valid when 'wr' is asserted.
        --

        delay_data : process(clk0)
        begin
            if clk0'event and clk0 = '1' then
                d_q <= EXT(d, rd_width * 2);

                if regd = '1' then
                    d_q_sel <= d_q;
                else
                    d_q_sel <= EXT(d, rd_width * 2);
                end if;
            end if;
        end process;

        --
        -- Here, DQ are generated coincident with the DQS pulses, on the
        -- assumption that the PCB traces for DQS introduce sufficient
        -- delay to meet DQS/DQ setup and hold times.
        --

        gen_ddrff : for i in 0 to rd_width - 1 generate
            U0 : FDDRRSE
                port map(
                    D0 => d_q_sel(i),
                    D1 => d_q_sel(i + rd_width),
                    C0 => clk180,
                    C1 => clk0,
                    CE => '1',
                    R  => '0',
                    S  => '0',
                    Q  => dq_o(i));
        end generate;
    end generate;

    --
    -- Generate output enables for DQ pins
    --

    process (rst, clk0)
    begin
        if rst = '1' then
            wr_q <= '0';
            wr_qq <= '0';
        elsif clk0'event and clk0 = '1' then
            wr_q <= wr;
            wr_qq <= wr_q;
        end if;
    end process;

    gen_dq_t : process (rst, clk180)
    begin
        if rst = '1' then
            dq_t <= (others => '1');
        elsif clk180'event and clk180 = '1' then
            dq_t <= oe_l;
        end if;
    end process;

    --
    -- Read data paths for various CAS latencies
    --
    -- The 'DQS' inputs provide clocks that capture the data, which is then
    -- recaptured on a clock edge that is determined by the 'tune' vector,
    -- on a per-bit basis, in order to allow for variations in flight time.
    --

    gen_delay : process(rst, clk0)
    begin
        if rst = '1' then
            delay <= (others => '0');
        elsif clk0'event and clk0 = '1' then
            delay <= delay(delay'high - 1 downto 0) & read_cmd;
        end if;
    end process;

    --
    -- Pipeline for tags (used only for reads)
    --

    gen_tag_pipe : process(clk0)
    begin
        if clk0'event and clk0 = '1' then
            tag_pipe(1 to tag_pipe'high) <= tag_pipe(0 to tag_pipe'high - 1);
            tag_pipe(0) <= tag;
        end if;
    end process;

    capture_0 : if capture_edge = 0 generate
        gen_reg : for i in 0 to dqs_width - 1 generate
            U0 : ddrcapture_reg
                generic map(
                    family => family,
                    width => 8)
                port map(
                    k0 => clk0,
                    k1 => clk180,
                    d => dq_i(8 * (i + 1) - 1 downto 8 * i),
                    q0 => q0_l0(8 * (i + 1) - 1 downto 8 * i),
                    q1 => q0_h0(8 * (i + 1) - 1 downto 8 * i));
        end generate;

        recapture : process(clk0)
        begin
            if clk0'event and clk0 = '1' then
                q0_l1 <= q_l0;
                q0_h1 <= q_h0;
            end if;
        end process;

        q <= EXT(q0_h1 & q0_l1, d_width);
    end generate;

    capture_1 : if capture_edge = 1 generate
        gen_reg : for i in 0 to dqs_width - 1 generate
            U0 : ddrcapture_reg
                generic map(
                    family => family,
                    width => 8)
                port map(
                    k0 => clk90,
                    k1 => clk270,
                    d => dq_i(8 * (i + 1) - 1 downto 8 * i),
                    q0 => q1_l0(8 * (i + 1) - 1 downto 8 * i),
                    q1 => q1_h0(8 * (i + 1) - 1 downto 8 * i));
        end generate;

        recapture : process(clk0)
        begin
            if clk0'event and clk0 = '1' then
                q1_l1 <= q1_l0;
                q1_h1 <= q1_h0;
            end if;
        end process;

        q <= EXT(q1_h1 & q1_l1, d_width);
    end generate;

    capture_2 : if capture_edge = 2 generate
        gen_reg : for i in 0 to dqs_width - 1 generate
            U0 : ddrcapture_reg
                generic map(
                    family => family,
                    width => 8)
                port map(
                    k0 => clk180,
                    k1 => clk0,
                    d => dq_i(8 * (i + 1) - 1 downto 8 * i),
                    q0 => q2_l0(8 * (i + 1) - 1 downto 8 * i),
                    q1 => q2_h0(8 * (i + 1) - 1 downto 8 * i));
        end generate;

        recapture : process(clk0)
        begin
            if clk0'event and clk0 = '1' then
                q2_l1 <= q2_l0;
            end if;
        end process;

        q <= EXT(q2_h0 & q2_l1, d_width);
    end generate;

    capture_3 : if capture_edge = 3 generate
        gen_reg : for i in 0 to dqs_width - 1 generate
            U0 : ddrcapture_reg
                generic map(
                    family => family,
                    width => 8)
                port map(
                    k0 => clk270,
                    k1 => clk90,
                    d => dq_i(8 * (i + 1) - 1 downto 8 * i),
                    q0 => q3_l0(8 * (i + 1) - 1 downto 8 * i),
                    q1 => q3_h0(8 * (i + 1) - 1 downto 8 * i));
        end generate;

        recapture180 : process(clk180)
        begin
            if clk180'event and clk180 = '1' then
                q3_l1 <= q3_l0;
            end if;
        end process;

        recapture0 : process(clk0)
        begin
            if clk0'event and clk0 = '1' then
                q3_l2 <= q3_l1;
                q3_h1 <= q3_h0;
            end if;
        end process;

        q <= EXT(q3_h1 & q3_l2, d_width);
    end generate;

    gen_qvalid : process(rst, clk0)
    begin
        if rst = '1' then
            qvalid <= '0';
            qtag <= (others => '-');
        elsif clk0'event and clk0 = '1' then
            if sr = '1' then
                qvalid <= '0';
            else
                if regd = '1' then
                    qvalid <= delay(capture_delay + 1);
                else
                    qvalid <= delay(capture_delay);
                end if;
            end if;

            if regd = '1' then
                qtag <= tag_pipe(capture_delay + 1);
            else
                qtag <= tag_pipe(capture_delay);
            end if;
        end if;
    end process;

end mixed;
