--
-- ddrsdram_data_dqs.vhd - DDR SDRAM data path that uses DQS signals to
--                         capture data from SDRAM.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

-- synopsys translate_off
library unisim;
use unisim.vcomponents.all;
library std;
use std.textio.all;
-- synopsys translate_on

library work;
use work.memif.all;

entity ddrsdram_data_dqs is
    generic(
        family      : in  integer;
        cas_latency : in  natural;
        flight_time : in  integer;
        rd_width    : in  natural;
        d_width     : in  natural;
        tag_width   : in  natural;
        dqs_delayed : in  boolean);
    port(
        clk0        : in  std_logic;
        clk90       : in  std_logic;
        clk180      : in  std_logic;
        clk270      : in  std_logic;
        rst         : in  std_logic;
        sr          : in  std_logic;
        regd        : in  std_logic;
        oe_l        : in  std_logic_vector(rd_width - 1 downto 0);
        wr          : in  std_logic;
        read_cmd    : in  std_logic;
        tune        : in  std_logic_vector(d_width * 2 - 1 downto 0);
        d           : in  std_logic_vector(d_width - 1 downto 0);
        tag         : in  std_logic_vector(tag_width - 1 downto 0);
        dm_i        : in  std_Logic_vector(rd_width / 8 - 1 downto 0);
        dqs_i       : in  std_Logic_vector(rd_width / 8 - 1 downto 0);
        dq_i        : in  std_Logic_vector(rd_width - 1 downto 0);
        dq_o        : out std_logic_vector(rd_width - 1 downto 0);
        dq_t        : out std_Logic_vector(rd_width - 1 downto 0);
        q           : out std_logic_vector(d_width - 1 downto 0);
        qtag        : out std_logic_vector(tag_width - 1 downto 0);
        qvalid      : out std_logic);
end;

architecture mixed of ddrsdram_data_dqs is

    -- synopsys translate_off
    function conv_string(
        val : in integer)
    return string is
        variable l : line;
        variable len : integer;
        variable s : string(1 to 64);
    begin
        write(l, val);
        len := l'length;
        read(l, s(1 to len));
        return s(1 to len);
    end;
    -- synopsys translate_on

    constant dqs_width : natural := rd_width / 8;

    constant capture_latency0 : natural := cas_latency * 2 + flight_time;
    constant capture_edge0 : natural := capture_latency0 mod 4;
    constant capture_delay0 : natural := (capture_latency0 + 8) / 4;
    constant capture_delay1 : natural := (capture_latency0 + 9) / 4;
    constant capture_delay2 : natural := (capture_latency0 + 10) / 4;
    constant capture_delay3 : natural := (capture_latency0 + 11) / 4;
    constant q0_extra_cycle : boolean := false;
    constant q1_extra_cycle : boolean := capture_edge0 = 1; 
    constant q2_extra_cycle : boolean := capture_edge0 = 1 or capture_edge0 = 2;
    constant q3_extra_cycle : boolean := capture_edge0 = 1 or capture_edge0 = 2 or capture_edge0 = 3;

    signal dqs_i_l : std_logic_vector(rd_width / 8 - 1 downto 0);
    signal dm_i_l : std_logic_vector(rd_width / 8 - 1 downto 0);

    signal d_q : std_logic_vector(rd_width * 2 - 1 downto 0);
    signal d_q_sel : std_logic_vector(rd_width * 2 - 1 downto 0);

    signal q_l, q_h : std_logic_vector(rd_width - 1 downto 0);

    signal q0_l0, q0_h0 : std_logic_vector(rd_width - 1 downto 0);
    signal q0_l1, q0_h1 : std_logic_vector(rd_width - 1 downto 0);
    signal q0_l2, q0_h2 : std_logic_vector(rd_width - 1 downto 0);
    signal q0_l3, q0_h3 : std_logic_vector(rd_width - 1 downto 0);
    signal q0 : std_logic_vector(d_width - 1 downto 0);

    signal q1_l0, q1_h0 : std_logic_vector(rd_width - 1 downto 0);
    signal q1_l1, q1_h1 : std_logic_vector(rd_width - 1 downto 0);
    signal q1_l2, q1_h2 : std_logic_vector(rd_width - 1 downto 0);
    signal q1_l3, q1_h3 : std_logic_vector(rd_width - 1 downto 0);
    signal q1 : std_logic_vector(d_width - 1 downto 0);

    signal q2_l0, q2_h0 : std_logic_vector(rd_width - 1 downto 0);
    signal q2_l1, q2_h1 : std_logic_vector(rd_width - 1 downto 0);
    signal q2_l2, q2_h2 : std_logic_vector(rd_width - 1 downto 0);
    signal q2_l3, q2_h3 : std_logic_vector(rd_width - 1 downto 0);
    signal q2 : std_logic_vector(d_width - 1 downto 0);

    signal q3_l0, q3_h0 : std_logic_vector(rd_width - 1 downto 0);
    signal q3_l1, q3_h1 : std_logic_vector(rd_width - 1 downto 0);
    signal q3_l2, q3_h2 : std_logic_vector(rd_width - 1 downto 0);
    signal q3_l3, q3_h3 : std_logic_vector(rd_width - 1 downto 0);
    signal q3 : std_logic_vector(d_width - 1 downto 0);

    signal q0sel : std_logic_vector(d_width - 1 downto 0);
    signal q1sel : std_logic_vector(d_width - 1 downto 0);
    signal q2sel : std_logic_vector(d_width - 1 downto 0);
    signal q3sel : std_logic_vector(d_width - 1 downto 0);

    signal delay : std_Logic_vector(7 downto 0);

    type tag_pipe_t is array(0 to 7) of std_logic_vector(tag_width - 1 downto 0);
    signal tag_pipe : tag_pipe_t;

    signal wr_q : std_logic;
    signal wr_qq : std_logic;

    signal trig0 : std_logic_vector(254 downto 0);
    signal trig_out : std_logic;

    attribute keep : string;
    attribute keep of trig_out : signal is "true";

    component FDDRRSE
        port(
            D0 : in  std_logic;
            D1 : in  std_logic;
            C0 : in  std_logic;
            C1 : in  std_logic;
            CE : in  std_logic;
            R  : in  std_logic;
            S  : in  std_logic;
            Q  : out std_logic);
    end component;

    component ddrcapture_reg
        generic(
            family     : in integer;
            width      : in natural);
        port(
            k0         : in  std_logic;                               -- Clock for 'q0'
            k1         : in  std_logic;                               -- Clock for 'q1'
            d          : in  std_logic_vector(width - 1 downto 0);    -- Flip-flop D input
            q0         : out std_logic_vector(width - 1 downto 0);    -- Flip-flop Q output for 'k0' clock
            q1         : out std_logic_vector(width - 1 downto 0));   -- Flip-flop Q output for 'k1' clock
    end component;

    --
    -- XST synthesis constraints, required to prevent optimization of
    -- flip-flops that would otherwise prevent IOB packing.
    --
    attribute equivalent_register_removal : string;
    attribute equivalent_register_removal of dq_o : signal is "no";
    attribute equivalent_register_removal of dq_t : signal is "no";
    attribute equivalent_register_removal of q_l : signal is "no";
    attribute equivalent_register_removal of q_h : signal is "no";
    -- Ensure that critical signals are packed into IOBs
    attribute iob : string;
    attribute iob of dq_t : signal is "false";
    attribute iob of dq_o : signal is "true";
    attribute iob of q_l : signal is "true";
    attribute iob of q_h : signal is "true";
    attribute iob of q0_l0 : signal is "false";
    attribute iob of q0_h0 : signal is "false";
    attribute iob of q1_l0 : signal is "false";
    attribute iob of q1_h0 : signal is "false";
    attribute iob of q2_l0 : signal is "false";
    attribute iob of q2_h0 : signal is "false";
    attribute iob of q3_l0 : signal is "false";
    attribute iob of q3_h0 : signal is "false";

begin

    -- synopsys translate_off
    assert cas_latency <= 8
        report "*** ddrsdram_data: illegal value " & conv_string(cas_latency) & " for cas_latency : greater than 8 half-cycles"
        severity failure;

    assert cas_latency >= 4
        report "*** ddrsdram_data: illegal value " & conv_string(cas_latency) & " for cas_latency : less than 4 half-cycles"
        severity failure;
    -- synopsys translate_on

    --
    -- Instantiate DDR registers for DQ pins
    --

    gen_d_phase90 : if not dqs_delayed generate
        --
        -- Write data path (independent of CAS latency)
        --
        -- 'd' is valid when 'wr' is asserted.
        --

        delay_data1 : process(clk0)
        begin
            if clk0'event and clk0 = '1' then
                d_q <= EXT(d, rd_width * 2);
            end if;
        end process;

        delay_data2 : process(clk270)
        begin
            if clk270'event and clk270 = '1' then
                if regd = '1' then
                    d_q_sel <= d_q;
                else
                    d_q_sel <= EXT(d, rd_width * 2);
                end if;
            end if;
        end process;

        --
        -- Here, DQ and DM are generated such that DQS falls in the middle
        -- of the valid window for DQ and DM. This is necessary when the
        -- PCB traces for DQS do not introduce significant delay.
        --

        gen_ddrff : for i in 0 to rd_width - 1 generate
            U0 : FDDRRSE
                port map(
                    D0 => d_q_sel(i),
                    D1 => d_q_sel(i + rd_width),
                    C0 => clk90,
                    C1 => clk270,
                    CE => '1',
                    R  => '0',
                    S  => '0',
                    Q  => dq_o(i));
        end generate;
    end generate;

    gen_d_phase180 : if dqs_delayed generate
        --
        -- Write data path (independent of CAS latency)
        --
        -- 'd' is valid when 'wr' is asserted.
        --

        delay_data : process(clk0)
        begin
            if clk0'event and clk0 = '1' then
                d_q <= EXT(d, rd_width * 2);

                if regd = '1' then
                    d_q_sel <= d_q;
                else
                    d_q_sel <= EXT(d, rd_width * 2);
                end if;
            end if;
        end process;

        --
        -- Here, DQ are generated coincident with the DQS pulses, on the
        -- assumption that the PCB traces for DQS introduce sufficient
        -- delay to meet DQS/DQ setup and hold times.
        --

        gen_ddrff : for i in 0 to rd_width - 1 generate
            U0 : FDDRRSE
                port map(
                    D0 => d_q_sel(i),
                    D1 => d_q_sel(i + rd_width),
                    C0 => clk180,
                    C1 => clk0,
                    CE => '1',
                    R  => '0',
                    S  => '0',
                    Q  => dq_o(i));
        end generate;
    end generate;

    --
    -- Generate output enables for DQ pins
    --

    process (rst, clk0)
    begin
        if rst = '1' then
            wr_q <= '0';
            wr_qq <= '0';
        elsif clk0'event and clk0 = '1' then
            wr_q <= wr;
            wr_qq <= wr_q;
        end if;
    end process;

    gen_dq_t : process (rst, clk180)
    begin
        if rst = '1' then
            dq_t <= (others => '1');
        elsif clk180'event and clk180 = '1' then
            dq_t <= oe_l;
        end if;
    end process;


    --
    -- Read data paths for various CAS latencies
    --
    -- One clock is used to capture data, which is then recaptured if necessary.
    --

    gen_delay : process(rst, clk0)
    begin
        if rst = '1' then
            delay <= (others => '0');
        elsif clk0'event and clk0 = '1' then
            delay <= delay(delay'high - 1 downto 0) & read_cmd;
        end if;
    end process;

    --
    -- Pipeline for tags (used only for reads)
    --

    gen_tag_pipe : process(clk0)
    begin
        if clk0'event and clk0 = '1' then
            tag_pipe(1 to tag_pipe'high) <= tag_pipe(0 to tag_pipe'high - 1);
            tag_pipe(0) <= tag;
        end if;
    end process;

    capture : for i in 0 to dqs_width - 1 generate
        -- Get inverted DQS
        dqs_i_l(i) <= not dqs_i(i);
        dm_i_l(i) <= not dm_i(i);

        by_8 : ddrcapture_reg
            generic map(
                family => family,
                width => 8)
            port map(
                k0 => dqs_i(i),
                k1 => dqs_i_l(i),
                d => dq_i(8 * (i + 1) - 1 downto 8 * i),
                q0 => q_l(8 * i + 7 downto 8 * i),
                q1 => q_h(8 * i + 7 downto 8 * i));
    end generate;
    
    recapture_q0_0 : process(clk0)
    begin
        if clk0'event and clk0 = '1' then
            q0_l0 <= q_l;
            q0_l1 <= q0_l0;
            q0_l2 <= q0_l1;
            q0_l3 <= q0_l2;
            q0_h1 <= q0_h0;
            q0_h2 <= q0_h1;
            q0_h3 <= q0_h2;
        end if;
    end process;

    recapture_q0_180 : process(clk180)
    begin
        if clk180'event and clk180 = '1' then
            q0_h0 <= q_h;
        end if;
    end process;

    q0 <= EXT(q0_h3 & q0_l3, d_width) when q0_extra_cycle else 
          EXT(q0_h2 & q0_l2, d_width);

    recapture_q1_0 : process(clk0)
    begin
        if clk0'event and clk0 = '1' then
            q1_l1 <= q1_l0;
            q1_l2 <= q1_l1;
            q1_l3 <= q1_l2;
            q1_h2 <= q1_h1;
            q1_h3 <= q1_h2;
        end if;
    end process;

    recapture_q1_90 : process(clk90)
    begin
        if clk90'event and clk90 = '1' then
            q1_l0 <= q_l;
        end if;
    end process;

    recapture_q1_180 : process(clk180)
    begin
        if clk180'event and clk180 = '1' then
            q1_h1 <= q1_h0;
        end if;
    end process;

    recapture_q1_270 : process(clk270)
    begin
        if clk270'event and clk270 = '1' then
            q1_h0 <= q_h;
        end if;
    end process;

    q1 <= EXT(q1_h3 & q1_l3, d_width) when q1_extra_cycle else
          EXT(q1_h2 & q1_l2, d_width);

    recapture_q2_0 : process(clk0)
    begin
        if clk0'event and clk0 = '1' then
            q2_l1 <= q2_l0;
            q2_l2 <= q2_l1;
            q2_l3 <= q2_l2;
            q2_h0 <= q_h;
            q2_h1 <= q2_h0;
            q2_h2 <= q2_h1;
        end if;
    end process;

    recapture_q2_180 : process(clk180)
    begin
        if clk180'event and clk180 = '1' then
            q2_l0 <= q_l;
        end if;
    end process;

    q2 <= EXT(q2_h2 & q2_l3, d_width) when q2_extra_cycle else
          EXT(q2_h1 & q2_l2, d_width);

    recapture_q3_0 : process(clk0)
    begin
        if clk0'event and clk0 = '1' then
            q3_l2 <= q3_l1;
            q3_l3 <= q3_l2;
            q3_h1 <= q3_h0;
            q3_h2 <= q3_h1;
        end if;
    end process;

    recapture_q3_90 : process(clk90)
    begin
        if clk90'event and clk90 = '1' then
            q3_h0 <= q_h;
        end if;
    end process;

    recapture_q3_180 : process(clk180)
    begin
        if clk180'event and clk180 = '1' then
            q3_l1 <= q3_l0;
        end if;
    end process;

    recapture_q3_270 : process(clk270)
    begin
        if clk270'event and clk270 = '1' then
            q3_l0 <= q_l;
        end if;
    end process;

    q3 <= EXT(q3_h2 & q3_l3, d_width) when q3_extra_cycle else
          EXT(q3_h1 & q3_l2, d_width);

    gen_qxsel : process(q0, q1, q2, q3)
    begin
        case capture_edge0 is
            when 0 =>
                q0sel <= q0;
                q1sel <= q1;
                q2sel <= q2;
                q3sel <= q3;
            when 1 =>
                q0sel <= q1;
                q1sel <= q2;
                q2sel <= q3;
                q3sel <= q0;
            when 2 =>
                q0sel <= q2;
                q1sel <= q3;
                q2sel <= q0;
                q3sel <= q1;
            when 3 =>
                q0sel <= q3;
                q1sel <= q0;
                q2sel <= q1;
                q3sel <= q2;
            when others =>
                q0sel <= (others => '-');
                q1sel <= (others => '-');
                q2sel <= (others => '-');
                q3sel <= (others => '-');
        end case;
    end process;

    gen_q : process(clk0)
        variable sel : std_logic_vector(1 downto 0);
    begin
        if clk0'event and clk0 = '1' then
            for i in q'range loop
                sel := tune(i * 2 + 1 downto i * 2);
                case sel is
                    when "00" =>
                        q(i) <= q0sel(i);
                    when "01" =>
                        q(i) <= q1sel(i);
                    when "10" =>
                        q(i) <= q2sel(i);
                    when "11" =>
                        q(i) <= q3sel(i);
                    when others =>
                        q(i) <= '-';
                end case;
            end loop;
        end if;
    end process;

    gen_qvalid : process(rst, clk0)
    begin
        if rst = '1' then
            qvalid <= '0';
            qtag <= (others => '-');
        elsif clk0'event and clk0 = '1' then
            if sr = '1' then
                qvalid <= '0';
            else
                if regd = '1' then
                    qvalid <= delay(capture_delay3 + 1);
                else
                    qvalid <= delay(capture_delay3);
                end if;
            end if;

            if regd = '1' then
                qtag <= tag_pipe(capture_delay3 + 1);
            else
                qtag <= tag_pipe(capture_delay3);
            end if;
        end if;
    end process;

end mixed;
