--
-- ddrsdram_init.vhd - DDR SDRAM memory interface, initialization state machine
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

-- synopsys translate_off
library std;
use std.textio.all;
-- synopsys translate_on

library work;
use work.memif.all;
use work.memif_def.all;

entity ddrsdram_init is
    generic(
        rowcol_width    : in    natural;
        bank_width      : in    natural;
        cas_latency     : in    natural;
        t_refresh       : in    natural;
        t_mrd           : in    natural;
        t_dllr          : in    natural;
        t_rp            : in    natural;
        t_rfc           : in    natural);
    port(
        clk             : in    std_logic;
        rst             : in    std_logic;
        sr              : in    std_logic;
        ref_done        : in    std_logic;
        ref_req         : out   std_logic;
        cke             : out   std_logic;
        done            : out   std_logic;
        command         : out   std_logic_vector(4 downto 0);
        addr            : out   std_logic_vector(rowcol_width + bank_width - 1 downto 0));
end;

architecture mixed of ddrsdram_init is

    -- synopsys translate_off
    function conv_string(
        val : in integer)
    return string is
        variable l : line;
        variable len : integer;
        variable s : string(1 to 64);
    begin
        write(l, val);
        len := l'length;
        read(l, s(1 to len));
        return s(1 to len);
    end;
    -- synopsys translate_on

    function choice(
        a, b : in integer;
        p : in boolean)
    return integer is
    begin
        if p then
            return b;
        else
            return a;
        end if;
    end;

    -- Order (4 downto 0) is (all, cs#, ras#, cas#, we#)
    constant DDRCMD_NOP    : std_logic_vector(4 downto 0) := "-1111"; -- Deselect
    constant DDRCMD_PRE    : std_logic_vector(4 downto 0) := "10010"; -- Precharge
    constant DDRCMD_REF    : std_logic_vector(4 downto 0) := "10001"; -- Refresh
    constant DDRCMD_MRS    : std_logic_vector(4 downto 0) := "10000"; -- Set mode register

    -- Number of cycles to wait after enabling clock to DDR SDRAM
    constant t_cke : natural := 5;

    signal modereg_val : std_logic_vector(6 downto 0);

    --
    -- Results in delay of approx. 0.25s @ 133MHz SDRAM clock for synthesis
    -- and a delay of approx 7.5us @ 133MHz SDRAM clock for simulation
    --
    constant pwr_on_timer_width : natural := choice(25, 10, memif_simulation);
    signal pwr_on_timer : std_logic_vector(pwr_on_timer_width - 1 downto 0);
    signal pwr_on_done : std_logic;

    type state1_type is (
        s1_power_on,
        s1_deselect,
        s1_precharge_all1,
        s1_emodereg,
        s1_modereg1,
        s1_precharge_all2,
        s1_refresh1,
        s1_refresh2,
        s1_modereg2,
        s1_done);
    signal state1, n_state1 : state1_type;
    signal i_addr, n_addr : std_logic_vector(14 downto 0);
    signal i_cke, n_cke : std_logic;
    signal i_command, n_command : std_logic_vector(4 downto 0);
    signal i_done, n_done : std_logic;
    signal en_refresh, n_en_refresh : std_logic;
    signal timer, n_timer : std_logic_vector(9 downto 0);

    signal refresh_timer : std_logic_vector(15 downto 0);

    type state2_type is (
        s2_idle,
        s2_request);
    signal state2, n_state2 : state2_type;
    signal i_ref_req, n_ref_req : std_logic;

    signal logic0, logic1 : std_logic;

begin

    -- synopsys translate_off
    assert t_dllr <= 1000
        report "*** ddrsdram_init: illegal value " & conv_string(t_dllr) & "for t_dllr : more than 1000"
        severity failure;

    assert t_rp >= 1
        report "*** ddrsdram_init: illegal value " & conv_string(t_rp) & "for t_rp : less than 1"
        severity failure;

    assert t_rp <= 31
        report "*** ddrsdram_init: illegal value " & conv_string(t_rp) & "for t_rp : more than 31"
        severity failure;

    assert t_rfc >= 1
        report "*** ddrsdram_init: illegal value " & conv_string(t_rfc) & "for t_rfc : less than 1"
        severity failure;

    assert t_rfc <= 31
        report "*** ddrsdram_init: illegal value " & conv_string(t_rfc) & "for t_rfc : more than 31"
        severity failure;
    -- synopsys translate_on

    logic0 <= '0';
    logic1 <= '1';

    modereg_val(2 downto 0) <= "001"; -- Burst length 2
    modereg_val(3) <= '0'; -- Sequential ordering
    modereg_val(6 downto 4) <= "101" when cas_latency = 3 else  -- CL1.5
                               "010" when cas_latency = 4 else  -- CL2
                               "110" when cas_latency = 5 else  -- CL2.5
                               "011" when cas_latency = 6 else  -- CL3
                               "100" when cas_latency = 8 else  -- CL4
                               "---";

    addr <= EXT(i_addr(14 downto 13), bank_width) & EXT(i_addr(12 downto 0), rowcol_width);
    cke <= i_cke;
    command <= i_command;
    done <= i_done;
    ref_req <= i_ref_req;

    --
    -- DDR SDRAM controller refresh state machine
    --

    gen_refresh_timer : process(rst, clk)
    begin
        if rst = '1' then
            refresh_timer <= (others => '0');
        elsif clk'event and clk = '1' then
            if en_refresh = '1' then
                if refresh_timer = t_refresh - 1 then
                    refresh_timer <= (others => '0');
                else
                    refresh_timer <= refresh_timer + 1;
                end if;
            else
                refresh_timer <= (others => '0');
            end if;
        end if;
    end process;

    gen_state2 : process(rst, clk)
    begin
        if rst = '1' then
            state2 <= s2_idle;
            i_ref_req <= '0';
        elsif clk'event and clk = '1' then
            if sr = '1' then
                state2 <= s2_idle;
                i_ref_req <= '0';
            else
                state2 <= n_state2;
                i_ref_req <= n_ref_req;
            end if;
        end if;
    end process;

    gen_n_state2 : process(
        state2,
        refresh_timer,
        ref_done)
    begin
        case state2 is
            when s2_idle =>
                if refresh_timer = t_refresh - 1 then
                    n_state2 <= s2_request;
                    n_ref_req <= '1';
                else
                    n_state2 <= s2_idle;
                    n_ref_req <= '0';
                end if;

            when s2_request =>
                if ref_done = '1' then
                    n_state2 <= s2_idle;
                    n_ref_req <= '0';
                else
                    n_state2 <= s2_request;
                    n_ref_req <= '1';
                end if;
        end case;
    end process;

    --
    -- Timer for timing the delay following reset/power-on
    --
    gen_pwr_on_timer : process(rst, clk)
    begin
        if rst = '1' then
            pwr_on_timer <= (others => '0');
            pwr_on_done <= '0';
        elsif clk'event and clk = '1' then
            if sr = '1' then
                pwr_on_timer <= (others => '0');
                pwr_on_done <= '0';
            else
                if pwr_on_done = '0' then
                    pwr_on_timer <= pwr_on_timer + 1;
                end if;

                if AND_reduce(pwr_on_timer) = '1' then
                    pwr_on_done <= '1';
                end if;
            end if;
        end if;
    end process;

    --
    -- DDR SDRAM initilization sequence state machine
    --

    gen_state1 : process(rst, clk)
    begin
        if rst = '1' then
            state1 <= s1_power_on;
            i_addr <= (others => '0');
            i_cke <= '0';
            i_command <= (others => '1');
            i_done <= '0';
            en_refresh <= '0';
            timer <= (others => '0');
        elsif clk'event and clk = '1' then
            if sr = '1' then
                state1 <= s1_power_on;
                i_addr <= (others => '0');
                i_cke <= '0';
                i_command <= (others => '1');
                i_done <= '0';
                en_refresh <= '0';
                timer <= (others => '0');
            else
                state1 <= n_state1;
                i_addr <= n_addr;
                i_cke <= n_cke;
                i_command <= n_command;
                i_done <= n_done;
                en_refresh <= n_en_refresh;
                timer <= n_timer;
            end if;
        end if;
    end process;

    gen_n_state1 : process(
        state1,
        modereg_val,
        pwr_on_done,
        timer)
    begin
        case state1 is
            when s1_power_on =>
                if pwr_on_done = '1' then
                    n_state1 <= s1_deselect;
                    n_addr <= (others => '-');
                    n_cke <= '1';
                    n_command <= DDRCMD_NOP;
                    n_done <= '0';
                    n_en_refresh <= '0';
                    n_timer <= (others => '0');
                else
                    n_state1 <= s1_power_on;
                    n_addr <= (others => '-');
                    n_cke <= '0';
                    n_command <= (others => '1');
                    n_done <= '0';
                    n_en_refresh <= '0';
                    n_timer <= (others => '-');
                end if;

            when s1_deselect =>
                if timer(2 downto 0) = t_cke then
                    -- Execute a 'precharge all' command
                    n_state1 <= s1_precharge_all1;
                    n_addr <= (10 => '1', others => '0');
                    n_cke <= '1';
                    n_command <= DDRCMD_PRE;
                    n_done <= '0';
                    n_en_refresh <= '0';
                    n_timer <= (others => '0');
                else
                    n_state1 <= s1_deselect;
                    n_cke <= '1';
                    n_addr <= (others => '-');
                    n_command <= DDRCMD_NOP;
                    n_done <= '0';
                    n_en_refresh <= '0';
                    n_timer <= timer + 1;
                end if;

            when s1_precharge_all1 =>
                if timer = t_rp then
                    -- Establish extended mode register
                    n_state1 <= s1_emodereg;
                    n_addr <= "010000000000000";
                    n_cke <= '1';
                    n_command <= DDRCMD_MRS;
                    n_done <= '0';
                    n_en_refresh <= '0';
                    n_timer <= (others => '0');
                else
                    n_state1 <= s1_precharge_all1;
                    n_addr <= (others => '-');
                    n_cke <= '1';
                    n_command <= DDRCMD_NOP;
                    n_done <= '0';
                    n_en_refresh <= '0';
                    n_timer <= timer + 1;
                end if;

            when s1_emodereg =>
                if timer = t_mrd then
                    -- Establish mode register and reset DLL
                    n_state1 <= s1_modereg1;
                    n_addr <= "00000010" & modereg_val;
                    n_cke <= '1';
                    n_command <= DDRCMD_MRS;
                    n_done <= '0';
                    n_en_refresh <= '0';
                    n_timer <= (others => '0');
                else
                    n_state1 <= s1_emodereg;
                    n_addr <= (others => '-');
                    n_cke <= '1';
                    n_command <= DDRCMD_NOP;
                    n_done <= '0';
                    n_en_refresh <= '0';
                    n_timer <= timer + 1;
                end if;

            when s1_modereg1 =>
                if timer = t_mrd then
                    -- Execute second 'precharge all' command
                    n_state1 <= s1_precharge_all2;
                    n_addr <= (10 => '1', others => '0');
                    n_cke <= '1';
                    n_command <= DDRCMD_PRE;
                    n_done <= '0';
                    n_en_refresh <= '0';
                    n_timer <= (others => '0');
                else
                    n_state1 <= s1_modereg1;
                    n_addr <= (others => '-');
                    n_cke <= '1';
                    n_command <= DDRCMD_NOP;
                    n_done <= '0';
                    n_en_refresh <= '0';
                    n_timer <= timer + 1;
                end if;

            when s1_precharge_all2 =>
                if timer = t_rp then
                    -- Execute first refresh cycle
                    n_state1 <= s1_refresh1;
                    n_addr <= (others => '-');
                    n_cke <= '1';
                    n_command <= DDRCMD_REF;
                    n_done <= '0';
                    n_en_refresh <= '0';
                    n_timer <= (others => '0');
                else
                    n_state1 <= s1_precharge_all2;
                    n_addr <= (others => '-');
                    n_cke <= '1';
                    n_command <= DDRCMD_NOP;
                    n_done <= '0';
                    n_en_refresh <= '0';
                    n_timer <= timer + 1;
                end if;

            when s1_refresh1 =>
                if timer = t_rfc then
                    -- Execute second refresh cycle
                    n_state1 <= s1_refresh2;
                    n_addr <= (others => '-');
                    n_cke <= '1';
                    n_command <= DDRCMD_REF;
                    n_done <= '0';
                    n_en_refresh <= '0';
                    n_timer <= (others => '0');
                else
                    n_state1 <= s1_refresh1;
                    n_addr <= (others => '-');
                    n_cke <= '1';
                    n_command <= DDRCMD_NOP;
                    n_done <= '0';
                    n_en_refresh <= '0';
                    n_timer <= timer + 1;
                end if;

            when s1_refresh2 =>
                if timer = t_rfc then
                    -- Reprogram mode register without resetting DLL
                    n_state1 <= s1_modereg2;
                    n_addr <= "00000000" & modereg_val;
                    n_cke <= '1';
                    n_command <= DDRCMD_MRS;
                    n_done <= '0';
                    n_en_refresh <= '1';
                    n_timer <= (others => '0');
                else
                    n_state1 <= s1_refresh2;
                    n_addr <= (others => '-');
                    n_cke <= '1';
                    n_command <= DDRCMD_NOP;
                    n_done <= '0';
                    n_en_refresh <= '0';
                    n_timer <= timer + 1;
                end if;

            when s1_modereg2 =>
                if timer = t_dllr then
                    -- Initialization completed
                    n_state1 <= s1_done;
                    n_addr <= (others => '-');
                    n_cke <= '1';
                    n_command <= DDRCMD_NOP;
                    n_done <= '0';
                    n_en_refresh <= '1';
                    n_timer <= (others => '-');
                else
                    n_state1 <= s1_modereg2;
                    n_addr <= (others => '-');
                    n_cke <= '1';
                    n_command <= DDRCMD_NOP;
                    n_done <= '0';
                    n_en_refresh <= '1';
                    n_timer <= timer + 1;
                end if;

            when s1_done =>
                -- Stay in this state for ever
                n_state1 <= s1_done;
                n_addr <= (others => '-');
                n_cke <= '1';
                n_command <= DDRCMD_NOP;
                n_done <= '1';
                n_en_refresh <= '1';
                n_timer <= (others => '-');
        end case;
    end process;

end mixed;
