--
-- memif_pkg.vhd - Package of memory interface components
--
-- MODEL SUPPORT
-- 
-- Model       ZBT SRAM       DDR-II SRAM    DDR SDRAM      DDR-II SDRAM
-- ----------------------------------------------------------------------
-- XRC         OK 0-80MHz     -              -              -
-- XRC-P       OK 0-80MHz     -              -              -
-- XRC-2L      OK 0-133MHz    -              -              -
-- XRC-2       OK 0-133MHz    -              -              -
-- DRC-II      -              TO DO          TO DO          -
-- WRC-II      -              -              TO DO          -
-- XPL         OK 0-133MHz    -              OK @133MHz     -
-- XP          -              OK @133MHz     OK @133MHz     -
-- XPI         -              TO DO          TO DO          -
-- XRC-4LX     OK 0-166MHz    -              -              -
-- XRC-4SX     OK 0-166MHz    -              -              -
-- XRC-4FX     -              -              -              OK @200MHz
-- XRC-5LX     -              -              -              OK @200-266MHz
-- XRC-5T1     -              TO DO          -              OK @200-266MHz
-- 

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

-- synopsys translate_off
library std;
use std.textio.all;
-- synopsys translate_on

package memif is

    constant memif_model_unknown : integer := -1;
    constant memif_model_admxrc : integer := 0;
    constant memif_model_admxrcp : integer := 1;
    constant memif_model_admxrc2l : integer := 2;
    constant memif_model_admxrc2 : integer := 3;
    constant memif_model_admxpl : integer := 4;
    constant memif_model_admxp : integer := 5;
    constant memif_model_adpdrc2 : integer := 6;
    constant memif_model_adpwrc2 : integer := 7;
    constant memif_model_adpxpi : integer := 8;
    constant memif_model_admxrc4ls : integer := 9;
    constant memif_model_admxrc4lx : integer := 10;
    constant memif_model_admxrc4sx : integer := 11;
    constant memif_model_admxrc4fx : integer := 12;
    constant memif_model_adpedev : integer := 13;
    constant memif_model_admxrc5lx : integer := 14;
    constant memif_model_admxrc5t1 : integer := 15;
    constant memif_model_xrm_ddr_c1 : integer := 100;

    constant memif_family_unknown : integer := -1;
    constant memif_family_auto : integer := 0;      -- Figure out family from model
    constant memif_family_virtex : integer := 1;    -- Virtex/Virtex-E/Virtex-EM
    constant memif_family_virtex2 : integer := 2;   -- Virtex-II
    constant memif_family_virtex2p : integer := 3;  -- Virtex-II Pro
    constant memif_family_virtex4 : integer := 4;   -- Virtex-4
    constant memif_family_virtex5 : integer := 5;   -- Virtex-5

    -- Type that represents a type of memory port
    type memif_port_type_t is (
        memif_port_ddrdram,
        memif_port_ddr2dram,
        memif_port_ddr2sram,
        memif_port_zbtsram);

    -- Type that describes a memory bank
    type memif_bank_t is
    record
        enable   : boolean;  -- 'true' if bank is physically present, otherwise false
        ra_width : natural;  -- Number of physical address wires in bank
        rc_width : natural;  -- Number of physical control wires in bank
        rd_width : natural;  -- Number of physical data wires in bank
    end record;

    constant memif_no_bank : memif_bank_t := (
        enable   => false,
        ra_width => 1, -- 1 instead of 0, to avoid 'std_logic_vector(-1 downto 0)'
        rc_width => 1,
        rd_width => 1);

    -- synopsys translate_off
    function memif_model_to_string(
        constant model : in integer)
    return string;
    -- synopsys translate_on

    function memif_resolve_family(
        constant family : in integer;
        constant model  : in integer)
    return integer;

    function memif_model_to_family(
        constant model : in integer)
    return integer;

    function memif_use_inverted_clocks(
        constant family : in integer;
        constant model  : in integer)
    return boolean;

    function memif_ddr2sram_capture_180deg(
        constant model  : in integer)
    return boolean;

    function memif_ddrsdram_max_rowcol_width(
        constant model : in integer)
    return integer;

    function memif_ddrsdram_max_bank_width(
        constant model : in integer)
    return integer;

    function memif_ddrsdram_max_num_phys_bank(
        constant model : in integer)
    return integer;

    function memif_ddrsdram_num_clock(
        constant model : in integer)
    return integer;

    function memif_ddrsdram_num_clock_en(
        constant model : in integer)
    return integer;

    function memif_ddrsdram_dq_before_dqs(
        constant model : in integer)
    return boolean;

    function memif_ddrsdram_dm_before_dqs(
        constant model : in integer)
    return boolean;

    function memif_ddrsdram_flight_time(
        constant model  : in integer)
    return natural;

    function memif_ddrsdram_use_dqs(
        constant model  : in integer)
    return boolean;

    function memif_ddrsdram_dynamic_tuning(
        constant model : in integer)
    return boolean;

    function memif_ddrsdram_tuning_vector(
        constant model : in integer;
        constant bank  : in natural)
    return std_logic_vector;

    function memif_ddrsdram_supports_dm(
        constant model : in integer)
    return boolean;

    function memif_ddr2sdram_num_clock(
        constant model : in integer)
    return integer;

    function memif_ddr2sdram_num_clock_en(
        constant model : in integer)
    return integer;

    function memif_ddr2sdram_max_num_phys_bank(
        constant model : in integer)
    return integer;

    function memif_ddr2sdram_max_bank_width(
        constant model : in integer)
    return integer;

    function memif_ddr2sdram_max_rowcol_width(
        constant model : in integer)
    return integer;

    component async_port
        generic(
            family     : in    integer;
            order      : in    natural;
            iwpfl      : in    natural;
            owpfl      : in    natural;
            orpel      : in    natural;
            addr_width : in    natural;
            data_width : in    natural;
            tag_width  : in    natural);
        port(
            rst        : in    std_logic;
            -- To/from primary clock domain
            pclk       : in    std_logic;
            psr        : in    std_logic;
            pce        : in    std_logic;
            pcw        : in    std_logic;
            pterm      : in    std_logic;
            padv       : in    std_logic;
            pwr        : in    std_logic;
            pa         : in    std_logic_vector(addr_width - 1 downto 0);
            pd         : in    std_logic_vector(data_width - 1 downto 0);
            pbe        : in    std_logic_vector(data_width / 8 - 1 downto 0);
            pq         : out   std_logic_vector(data_width - 1 downto 0);
            pwf        : out   std_logic;
            pwpf       : out   std_logic;
            pre        : out   std_logic;
            prpe       : out   std_logic;
            -- To/from memory interface
            sclk       : in    std_logic;
            sce        : out   std_logic;
            stag       : out   std_logic_vector(tag_width - 1 downto 0);
            sw         : out   std_logic;
            sa         : out   std_logic_vector(addr_width - 1 downto 0);
            sd         : out   std_logic_vector(data_width - 1 downto 0);
            sbe        : out   std_logic_vector(data_width / 8 - 1 downto 0);
            sready     : in    std_logic;
            svalid     : in    std_logic;
            sq         : in    std_logic_vector(data_width - 1 downto 0);
            sqtag      : in    std_logic_vector(tag_width - 1 downto 0);
            -- To/from memory interface arbiter
            sreq       : out   std_logic;
            -- Should be driven with constant values
            tag_base   : in    std_logic_vector(tag_width - 1 downto 0);
            tag_mask   : in    std_logic_vector(tag_width - 1 downto 0));
    end component;

    component ddrsdram_port
        generic(
            family          : in    integer := memif_family_auto;
            model           : in    integer;
            bank_index      : in    natural;
            byte_enables    : in    boolean := true;
            ra_width        : in    natural;
            rc_width        : in    natural;
            rd_width        : in    natural;
            a_width         : in    natural;
            d_width         : in    natural;
            tag_width       : in    natural;
            cas_latency     : in    natural;
            t_refresh       : in    natural;                                       -- Refresh period, in cycles
            --
            -- The following represent delays between the issue of non-NOP commands.
            -- A value of 1 cycle means that the next command can follow the current
            -- command with no intervening NOP commands.
            -- A value of 0 cycles is illegal.
            --
            t_mrd           : in    natural;                                       -- Min. delay between mode reg. cmd and another cmd, in cycles
            t_dllr          : in    natural;                                       -- Min. delay between DLL reset and another cmd, in cycles
            t_rp            : in    natural;                                       -- Min. delay between precharge cmd and another cmd, in cycles
            t_rfc           : in    natural;                                       -- Min. delay between refresh cmd and another cmd, in cycles
            t_act           : in    natural;                                       -- Min. delay between activation cmd and read/write cmd, in cycles
            t_wtr           : in    natural;                                       -- Min. delay between write cmd and read cmd, in cycles
            t_rtw           : in    natural;                                       -- Min. delay between read cmd and write cmd, in cycles
            t_rtp           : in    natural;                                       -- Min. delay between read cmd and precharge cmd, in cycles
            t_wtp           : in    natural;                                       -- Min. delay between write cmd and precharge cmd, in cycles
            t_ras           : in    natural);                                      -- Min. bank open time, in cycles
        port(
            -- To/from application
            clk0            : in    std_logic;
            clk90           : in    std_logic;
            clk180          : in    std_logic;
            clk270          : in    std_logic;
            rst             : in    std_logic;                                  -- asynchronous reset
            sr              : in    std_logic;                                  -- synchronous reset
            ce              : in    std_logic;                                  -- ignored unless 'ready'
            tag             : in    std_logic_vector(tag_width - 1 downto 0);   -- qualified by 'ce and not w'
            w               : in    std_logic;                                  -- qualified by 'ce'
            a               : in    std_logic_vector(a_width - 1 downto 0);     -- qualified by 'ce'
            d               : in    std_logic_vector(d_width - 1 downto 0);     -- qualified by 'ce and w'
            be              : in    std_logic_vector(d_width / 8 - 1 downto 0); -- qualified by 'ce and w'
            ready           : out   std_logic;                                  -- qualifies 'ce'
            valid           : out   std_logic;                                  -- indicates 'q' and 'qtag' valid
            q               : out   std_logic_vector(d_width - 1 downto 0);     -- qualified by 'valid'
            qtag            : out   std_logic_vector(tag_width - 1 downto 0);   -- qualified by 'valid'
            -- Sideband control signals, must be stable during operation
            regd            : in    std_logic;                                  -- 0 => unbuffered, 1 => registered
            x4              : in    std_logic;                                  -- 0 => normal, 1 => DM functions as DQS (x4 devices on DIMM)
            row             : in    std_logic_vector(1 downto 0);               -- row addr width: 00 => 12, ... , 11 => 15
            col             : in    std_logic_vector(1 downto 0);               -- col addr width: 00 => row - 3, ... , 11 => row
            bank            : in    std_logic_vector(1 downto 0);               -- bank addr width: 00 => 0, ... , 11 => 3
            pbank           : in    std_logic_vector(1 downto 0);               -- num. phys banks: 00 => 1, ... , 11 => 8
            -- To/from memories
            ra              : out   std_logic_vector(ra_width - 1 downto 0);
            rc              : inout std_logic_vector(rc_width - 1 downto 0);
            rd              : inout std_logic_vector(rd_width - 1 downto 0));
    end component;

    component ddr2sdram_port
        generic(
            auto_train      : in    boolean := true;
            family          : in    integer := memif_family_auto;
            model           : in    integer;
            bank_index      : in    natural;
            ra_width        : in    natural;
            rc_width        : in    natural;
            rd_width        : in    natural;
            a_width         : in    natural;
            d_width         : in    natural;
            tag_width       : in    natural;
            t_refresh       : in    natural;
            t_mrd           : in    natural;
            t_dllr          : in    natural;
            t_rp            : in    natural;
            t_rfc           : in    natural;
            t_act           : in    natural;
            t_wtr           : in    natural;
            t_rtw           : in    natural;
            t_rtp           : in    natural;
            t_wtp           : in    natural;
            t_ras           : in    natural);
        port(
            -- To/from DDR-II SDRAM infrastructure
            clk0            : in    std_logic;      -- Interface clock
            clk45           : in    std_logic;      -- As 'clk2x90' but divided by 2
            clk2x0          : in    std_logic;      -- As 'clk0' but double frequency
            clk2x90         : in    std_logic;      -- As 'clk2x0' but -90 deg. phase shift
            rst             : in    std_logic;      -- Reset
            -- To/from application
            sr              : in    std_logic;                                   -- Synchronous reset
            ce              : in    std_logic;                                   -- Command enable, qualified by 'ready'
            tag             : in    std_logic_vector(tag_width - 1 downto 0);    -- Tag, qualified by 'ce and not w')
            w               : in    std_logic;                                   -- 1 => write, qualified by 'ce')
            a               : in    std_logic_vector(a_width - 1 downto 0);      -- Address, qualified by 'ce'
            d               : in    std_logic_vector(d_width - 1 downto 0);      -- Write data, qualified by 'ce and w'
            be              : in    std_logic_vector(d_width / 8 - 1 downto 0);  -- Write byte enables, qualified by 'ce and w'
            q               : out   std_logic_vector(d_width - 1 downto 0);      -- Read data, qualified by 'valid'
            qtag            : out   std_logic_vector(tag_width - 1 downto 0);    -- Read tag, qualified by 'valid'
            valid           : out   std_logic;                                   -- 1 => 'q' and 'qtag' are valid
            ready           : out   std_logic;                                   -- When low, indicates that 'ce' is ignored
            -- Sideband control signals (inputs must be stable during operation)
            row             : in    std_logic_vector(1 downto 0);                -- row addr width: 00 => 12, ... , 11 => 15
            col             : in    std_logic_vector(1 downto 0);                -- col addr width: 00 => row - 4, ... , 11 => row - 1
            bank            : in    std_logic_vector(1 downto 0);                -- bank addr width: 00 => 0, ... , 11 => 3
            pbank           : in    std_logic_vector(1 downto 0);                -- num. phys banks: 00 => 1, ... , 11 => 8
            trained         : out   std_logic;                                   -- 1 => training was successful
            -- To/from memories
            ra              : out   std_logic_vector(ra_width - 1 downto 0);
            rc              : inout std_logic_vector(rc_width - 1 downto 0);
            rd              : inout std_logic_vector(rd_width - 1 downto 0));
    end component;

    component ddr2sdram_port_mig
        generic(
            family          : in    integer := memif_family_auto;
            model           : in    integer;
            ra_width        : in    natural;
            rc_width        : in    natural;
            rd_width        : in    natural;
            a_width         : in    natural;
            d_width         : in    natural;
            tag_width       : in    natural);
        port(
            -- To/from DDR-II SDRAM infrastructure
            clk0            : in    std_logic;      -- DDR SDRAM clock
            clk90           : in    std_logic;      -- As 'clk0' but -90 deg. phase shift
            clkdiv0         : in    std_logic;      -- System clock ('clk0' divided by 2)
            clkdiv90        : in    std_logic;      -- As 'clkdiv0' but -90 deg. phase shift
            clkdiv90ub      : in    std_logic;      -- As 'clkdiv90' but unbuffered
            clkdiv180ub     : in    std_logic;      -- As 'clkdiv90ub' but further -90 deg. phase
            rst             : in    std_logic;      -- Reset
            rst90           : in    std_logic;      -- Reset
            idelayctrl_rdy  : in    std_logic;      -- IDELAYCTRL ready
            -- To/from application
            sr              : in    std_logic;                                   -- Synchronous reset
            ce              : in    std_logic;                                   -- Command enable, qualified by 'ready'
            tag             : in    std_logic_vector(tag_width - 1 downto 0);    -- Tag, qualified by 'ce and not w')
            w               : in    std_logic;                                   -- 1 => write, qualified by 'ce')
            a               : in    std_logic_vector(a_width - 1 downto 0);      -- Address, qualified by 'ce'
            d               : in    std_logic_vector(d_width - 1 downto 0);      -- Write data, qualified by 'ce and w'
            be              : in    std_logic_vector(d_width / 8 - 1 downto 0);  -- Write byte enables, qualified by 'ce and w'
            q               : out   std_logic_vector(d_width - 1 downto 0);      -- Read data, qualified by 'valid'
            qtag            : out   std_logic_vector(tag_width - 1 downto 0);    -- Read tag, qualified by 'valid'
            valid           : out   std_logic;                                   -- 1 => 'q' and 'qtag' are valid
            ready           : out   std_logic;                                   -- When low, indicates that 'ce' is ignored
            -- Sideband control signals (inputs must be stable during operation)
            trained         : out   std_logic;                                   -- 1 => training was successful
            -- To/from memories
            ra              : out   std_logic_vector(ra_width - 1 downto 0);
            rc              : inout std_logic_vector(rc_width - 1 downto 0);
            rd              : inout std_logic_vector(rd_width - 1 downto 0));
    end component;

    component ddr2sram_port
        generic(
            family          : in    integer := memif_family_auto;
            model           : in    integer;
            ra_width        : in    natural;
            rc_width        : in    natural;
            rd_width        : in    natural;
            a_width         : in    natural;
            d_width         : in    natural;
            tag_width       : in    natural);
        port(
            clk0            : in    std_logic;                                   -- Clock
            clk90           : in    std_logic;
            clk180          : in    std_logic;
            clk270          : in    std_logic;
            rst             : in    std_logic;                                   -- Aynchronous reset
            sr              : in    std_logic;                                   -- Synchronous reset
            ce              : in    std_logic;                                   -- Command enable, qualified by 'ready'
            tag             : in    std_logic_vector(tag_width - 1 downto 0);    -- Tag, qualified by 'ce and not w')
            w               : in    std_logic;                                   -- 1 => write, qualified by 'ce')
            a               : in    std_logic_vector(a_width - 1 downto 0);      -- Address, qualified by 'ce'
            d               : in    std_logic_vector(d_width - 1 downto 0);      -- Write data, qualified by 'ce and w'
            be              : in    std_logic_vector(d_width / 8 - 1 downto 0);  -- Write byte enables, qualified by 'ce and w'
            q               : out   std_logic_vector(d_width - 1 downto 0);      -- Read data, qualified by 'valid'
            qtag            : out   std_logic_vector(tag_width - 1 downto 0);    -- Read tag, qualified by 'valid'
            valid           : out   std_logic;                                   -- 1 => 'q' and 'qtag' are valid
            ready           : out   std_logic;                                   -- When low, indicates that 'ce' is ignored
            -- Sideband control signals, must be stable during operation
            burst_len       : in    std_logic;                                   -- Supported burst length: 0 => 2 only, 1 => 2 & 4
            dll_off         : in    std_logic;                                   -- Disables DLL in SSRAM device when high
            -- To/from memories
            ra              : out   std_logic_vector(ra_width - 1 downto 0);
            rc              : inout std_logic_vector(rc_width - 1 downto 0);
            rd              : inout std_logic_vector(rd_width - 1 downto 0));
    end component;

    component port_mux
        generic(
            order           : in    natural;                                       -- Must be log2(in_width / out_width)
            in_width        : in    natural;                                       -- Unmultiplexed data width
            out_width       : in    natural);                                      -- Multiplexed segment width
        port(
            rst             : in    std_logic;                                     -- Asychronous reset
            clk             : in    std_logic;                                     -- Clock
            init            : in    std_logic;                                     -- Pulse to load 'addr'
            addr            : in    std_logic_vector(order - 1 downto 0);          -- Selects first segment of a 'din'
            adv             : in    std_logic;                                     -- Advances 'dout' to next segment
            din             : in    std_logic_vector(in_width - 1 downto 0);       -- Unmultiplexed data
            dout            : out   std_logic_vector(out_width - 1 downto 0);      -- Multiplexed segment
            last            : out   std_logic);                                    -- Asserted when 'dout' is last segment of 'din'
    end component;

    component port_repl
        generic(
            order           : in    natural;
            in_width        : in    natural;
            out_width       : in    natural;
            partial         : in    boolean);
        port(
            rst             : in    std_logic;
            clk             : in    std_logic;
            init            : in    std_logic;
            addr            : in    std_logic_vector(order - 1 downto 0);
            wr              : in    std_logic;
            last            : in    std_logic;
            din             : in    std_logic_vector(in_width - 1 downto 0);
            bein            : in    std_logic_vector(in_width / 8 - 1 downto 0);
            dout            : out   std_logic_vector(out_width - 1 downto 0);      -- Qual. by 'valid'
            beout           : out   std_logic_vector(out_width / 8 - 1 downto 0);  -- Qual. by 'valid'
            valid           : out   std_logic;
            final           : out   std_logic);                                    -- Qual. by 'valid'
    end component;

    component pulse_sync
        port(
            ar : in  std_logic;
            ik : in  std_logic;
            i  : in  std_logic;
            ok : in  std_logic;
            o  : out std_logic);
    end component;

    component spd_i2c
        generic(
            scldiv          : in    natural;
            enable_write    : in    boolean := false);
        port(
            -- Client interface
            clk             : in    std_logic;                                     -- Clock
            rst             : in    std_logic;                                     -- Asynchronous reset
            sr              : in    std_logic;                                     -- Synchronous reset
            ce              : in    std_logic;                                     -- Command enable, initiates operation
            wr              : in    std_logic;                                     -- Qualified by 'ce', '1' => write
            sa              : in    std_logic_vector(2 downto 0);                  -- Slot address
            a               : in    std_logic_vector(7 downto 0);                  -- Location in EEPROM to access
            d               : in    std_logic_vector(7 downto 0);                  -- Word to be written to EEPROM
            ack             : out   std_logic;                                     -- Pulsed to indicate read/write finished
            busy            : out   std_logic;                                     -- '1' => 'ce' is ignored
            q               : out   std_logic_vector(7 downto 0);                  -- Word read from EEPROM, qual. by 'ack'
            err             : out   std_logic;                                     -- '1' => operation failed, qual. by 'ack'
            -- To/from EEPROM pins
            sda_i           : in    std_logic;                                     -- Serial data from EEPROM
            sda_o           : out   std_logic;                                     -- Serial data to EEPROM
            scl_i           : in    std_logic;                                     -- Serial clock from EEPROM
            scl_o           : out   std_logic);                                    -- Serial clock to EEPROM
    end component;
            
    component zbtsram_port
        generic(
            family          : in    integer := memif_family_auto;
            model           : in    integer;
            ra_width        : in    natural;
            rd_width        : in    natural;
            rc_width        : in    natural;
            a_width         : in    natural;
            d_width         : in    natural;
            tag_width       : in    natural);
        port(
            clk0            : in    std_logic;
            rst             : in    std_logic;
            sr              : in    std_logic;
            pipeline        : in    std_logic;
            ce              : in    std_logic;
            tag             : in    std_logic_vector(tag_width - 1 downto 0);
            w               : in    std_logic;
            a               : in    std_logic_vector(a_width - 1 downto 0);
            d               : in    std_logic_vector(d_width - 1 downto 0);
            be              : in    std_logic_vector(d_width / 8 - 1 downto 0);
            q               : out   std_logic_vector(d_width - 1 downto 0);
            qtag            : out   std_logic_vector(tag_width - 1 downto 0);
            valid           : out   std_logic;
            ready           : out   std_logic;
            ra              : out   std_logic_vector(ra_width - 1 downto 0);
            rc              : inout std_logic_vector(rc_width - 1 downto 0);
            rd              : inout std_logic_vector(rd_width - 1 downto 0));
    end component;

end;

package body memif is

    -- synopsys translate_off
    function conv_string(
        val : in integer)
    return string is
        variable l : line;
        variable len : integer;
        variable s : string(1 to 64);
    begin
        write(l, val);
        len := l'length;
        read(l, s(1 to len));
        return s(1 to len);
    end;
    -- synopsys translate_on

    -- synopsys translate_off
    function memif_model_to_string(
        constant model : in integer)
    return string is
    begin
        case model is
            when memif_model_admxrc =>
                return "ADM-XRC";
            when memif_model_admxrcp =>
                return "ADM-XRC-P";
            when memif_model_admxrc2l =>
                return "ADM-XRC-II-Lite";
            when memif_model_admxrc2 =>
                return "ADM-XRC-II";
            when memif_model_admxpl =>
                return "ADM-XPL";
            when memif_model_admxp =>
                return "ADM-XP";
            when memif_model_adpdrc2 =>
                return "ADP-DRC-II";
            when memif_model_adpwrc2 =>
                return "ADP-WRC-II";
            when memif_model_adpxpi =>
                return "ADP-XPI";
            when memif_model_admxrc4ls =>
                return "ADM-XRC-4LS";
            when memif_model_admxrc4lx =>
                return "ADM-XRC-4LX";
            when memif_model_admxrc4sx =>
                return "ADM-XRC-4SX";
            when memif_model_admxrc4fx =>
                return "ADM-XRC-4FX";
            when memif_model_adpedev =>
                return "ADPE-DEV";
            when memif_model_admxrc5lx =>
                return "ADM-XRC-5LX";
            when memif_model_admxrc5t1 =>
                return "ADM-XRC-5T1";
            when others =>
                return conv_string(model);
        end case;
    end;
    -- synopsys translate_on

    function memif_resolve_family(
        constant family : in integer;
        constant model  : in integer)
    return integer is
    begin
        case family is
            when memif_family_auto =>
                return memif_model_to_family(model);
            when others =>
                return family;
        end case;
    end;

    function memif_model_to_family(
        constant model : in integer)
    return integer is
    begin
        case model is
            when memif_model_admxrc | memif_model_admxrcp =>
                return memif_family_virtex;
            when memif_model_admxrc2l | memif_model_admxrc2 | memif_model_adpdrc2 | memif_model_adpwrc2 =>
                return memif_family_virtex2;
            when memif_model_admxpl | memif_model_admxp | memif_model_adpxpi =>
                return memif_family_virtex2p;
            when memif_model_admxrc4ls | memif_model_admxrc4lx | memif_model_admxrc4sx | memif_model_admxrc4fx | memif_model_adpedev =>
                return memif_family_virtex4;
            when memif_model_admxrc5lx | memif_model_admxrc5t1 =>
                return memif_family_virtex5;
            when memif_model_xrm_ddr_c1 =>
                -- synopsys translate_off
                assert false
                    report "*** memif_model_to_family: can't determine model with XRM-DDR-C1"
                    severity failure;
                -- synopsys translate_on
                return memif_family_unknown;
            when others =>
                -- synopsys translate_off
                assert false
                    report "*** memif_model_to_family: unknown model ( " & memif_model_to_string(model) & ")"
                    severity failure;
                -- synopsys translate_on
                return memif_family_unknown;
        end case;
    end;

    function memif_use_inverted_clocks(
        constant family : in integer;
        constant model  : in integer)
    return boolean is
    begin
        case family is
            when memif_family_auto =>
                return memif_use_inverted_clocks(memif_model_to_family(model), model);
            when memif_family_virtex =>
                return true;
            when memif_family_virtex2 =>
                return true;
            when memif_family_virtex2p =>
                return false;
            when memif_family_virtex4 =>
                return true;
            when memif_family_virtex5 =>
                return true;
            when others =>
                -- synopsys translate_off
                assert false
                    report "*** memif_use_inverted_clocks: unknown family ( " & conv_string(family) & ")"
                    severity failure;
                -- synopsys translate_on
                return false;
        end case;
    end;

    function memif_ddr2sram_capture_180deg(
        constant model : in integer)
    return boolean is
    begin
        case model is
            when memif_model_adpdrc2 =>
                return true;
            when memif_model_admxp =>
                return true;
            when memif_model_adpxpi =>
                return true;
            when others =>
                -- synopsys translate_off
                assert false
                    report "*** memif_model_ddr2sram_capture_180deg: unknown model ( " & memif_model_to_string(model) & ")"
                    severity failure;
                -- synopsys translate_on
                return false;
        end case;
    end;

    function memif_ddrsdram_max_rowcol_width(
        constant model : in integer)
    return integer is
    begin
        case model is
            when memif_model_adpdrc2 | memif_model_adpwrc2 =>
                return 13;
            when memif_model_admxpl | memif_model_admxp =>
                return 13;
            when memif_model_adpxpi =>
                return 16;
            when memif_model_xrm_ddr_c1 =>
                return 14;
            when others =>
                -- synopsys translate_off
                assert false
                    report "*** memif_ddrsdram_max_rowcol_width: unknown model ( " & memif_model_to_string(model) & ")"
                    severity failure;
                -- synopsys translate_on
                return 0;
        end case;
    end;

    function memif_ddrsdram_max_bank_width(
        constant model : in integer)
    return integer is
    begin
        case model is
            when memif_model_adpdrc2 =>
                return 2;
            when memif_model_adpwrc2 =>
                return 2;
            when memif_model_admxpl | memif_model_admxp =>
                return 2;
            when memif_model_adpxpi =>
                return 3;
            when memif_model_xrm_ddr_c1 =>
                return 2;
            when others =>
                -- synopsys translate_off
                assert false
                    report "*** memif_ddrsdram_max_bank_width: unknown model ( " & memif_model_to_string(model) & ")"
                    severity failure;
                -- synopsys translate_on
                return 0;
        end case;
    end;

    function memif_ddrsdram_max_num_phys_bank(
        constant model : in integer)
    return integer is
    begin
        case model is
            when memif_model_adpdrc2 =>
                return 2;
            when memif_model_adpwrc2 =>
                return 1;
            when memif_model_admxpl | memif_model_admxp =>
                return 1;
            when memif_model_adpxpi =>
                return 4;
            when memif_model_xrm_ddr_c1 =>
                return 2;
            when others =>
                -- synopsys translate_off
                assert false
                    report "*** memif_ddrsdram_max_num_phys_bank: unknown model ( " & memif_model_to_string(model) & ")"
                    severity failure;
                -- synopsys translate_on
                return 0;
        end case;
    end;

    function memif_ddrsdram_num_clock(
        constant model : in integer)
    return integer is
    begin
        case model is
            when memif_model_adpdrc2 =>
                return 3;
            when memif_model_adpwrc2 =>
                return 1;
            when memif_model_admxpl | memif_model_admxp =>
                return 1;
            when memif_model_adpxpi =>
                return 3;
            when memif_model_xrm_ddr_c1 =>
                return 1;
            when others =>
                -- synopsys translate_off
                assert false
                    report "*** memif_ddrsdram_num_clock: unknown model ( " & memif_model_to_string(model) & ")"
                    severity failure;
                -- synopsys translate_on
                return 0;
        end case;
    end;

    function memif_ddrsdram_num_clock_en(
        constant model : in integer)
    return integer is
    begin
        case model is
            when memif_model_adpdrc2 =>
                return 2;
            when memif_model_adpwrc2 =>
                return 1;
            when memif_model_admxpl | memif_model_admxp =>
                return 1;
            when memif_model_adpxpi =>
                return 2;
            when memif_model_xrm_ddr_c1 =>
                return 1;
            when others =>
                -- synopsys translate_off
                assert false
                    report "*** memif_ddrsdram_num_clock_en: unknown model ( " & memif_model_to_string(model) & ")"
                    severity failure;
                -- synopsys translate_on
                return 0;
        end case;
    end;

    function memif_ddrsdram_dq_before_dqs(
        constant model : in integer)
    return boolean is
    begin
        case model is
            when memif_model_adpdrc2 =>
                return true;
            when memif_model_adpwrc2 =>
                return false;
            when memif_model_admxpl | memif_model_admxp =>
                return false;
            when memif_model_adpxpi =>
                return false;
            when memif_model_xrm_ddr_c1 =>
                return false;
            when others =>
                -- synopsys translate_off
                assert false
                    report "*** memif_model_to_dqs_delayed: unknown model ( " & memif_model_to_string(model) & ")"
                    severity failure;
                -- synopsys translate_on
                return false;
        end case;
    end;

    function memif_ddrsdram_dm_before_dqs(
        constant model : in integer)
    return boolean is
    begin
        case model is
            when memif_model_adpdrc2 =>
                return true;
            when memif_model_adpwrc2 =>
                return false;
            when memif_model_admxpl | memif_model_admxp =>
                return false;
            when memif_model_adpxpi =>
                return false;
            when memif_model_xrm_ddr_c1 =>
                return false;
            when others =>
                -- synopsys translate_off
                assert false
                    report "*** memif_model_to_dqs_delayed: unknown model ( " & memif_model_to_string(model) & ")"
                    severity failure;
                -- synopsys translate_on
                return false;
        end case;
    end;

    function memif_ddrsdram_flight_time(
        constant model  : in integer)
    return natural is
    begin
        case model is
            when memif_model_admxpl =>
                return 0;
            when memif_model_admxp =>
                return 1;
            when memif_model_adpdrc2 =>
                return 3;
            when memif_model_adpwrc2 =>
                return 3;
            when memif_model_adpxpi =>
                return 4; -- for 125/133MHz
            when others =>
                -- synopsys translate_off
                assert false
                    report "*** memif_ddrsdram_flight_time: unknown model ( " & memif_model_to_string(model) & ")"
                    severity failure;
                -- synopsys translate_on
                return 0;
        end case;
    end;

    function memif_ddrsdram_use_dqs(
        constant model  : in integer)
    return boolean is
    begin
        case model is
            when memif_model_admxpl | memif_model_adpwrc2 | memif_model_admxp =>
                return false;
            when memif_model_adpdrc2 | memif_model_adpxpi =>
                return true;
            when others =>
                -- synopsys translate_off
                assert false
                    report "*** memif_ddrsdram_use_dqs: unknown model ( " & memif_model_to_string(model) & ")"
                    severity failure;
                -- synopsys translate_on
                return false;
        end case;
    end;

    function memif_ddrsdram_use_IDELAY_for_dqs(
        constant family : in integer)
    return boolean is
    begin
        case family is
            when memif_family_virtex =>
                return false;
            when memif_family_virtex2 =>
                return false;
            when memif_family_virtex2p =>
                return false;
            when memif_family_virtex4 =>
                return true;
            when memif_family_virtex5 =>
                return true;
            when others =>
                -- synopsys translate_off
                assert false
                    report "*** memif_ddrsdram_use_IDELAY_for_dqs: unknown family ( " & conv_string(family) & ")"
                    severity failure;
                -- synopsys translate_on
                return false;
        end case;
    end;

    function memif_ddrsdram_dynamic_tuning(
        constant model : in integer)
    return boolean is
    begin
        case model is
            when memif_model_admxpl | memif_model_admxp =>
                return false;
            when memif_model_adpdrc2 =>
                return true;
            when memif_model_adpwrc2 =>
                return true;
            when memif_model_adpxpi =>
                return true;
            when others =>
                -- synopsys translate_off
                assert false
                    report "*** memif_ddrsdram_dynamic_tuning: unknown model ( " & memif_model_to_string(model) & ")"
                    severity failure;
                -- synopsys translate_on
                return false;
        end case;
    end;

    function memif_ddrsdram_tuning_vector(
        constant model : in integer;
        constant bank  : in natural)
    return std_logic_vector is
    begin
        case model is
            when memif_model_adpdrc2 =>
                case bank is
                    when 0 =>
                        return X"55555555" & X"55555555" & X"55555555" & X"55555555" & X"55555555" & X"55555555" & X"55555555" & X"55555555";
                    when 1 =>
                        return X"55555555" & X"55555555" & X"55555555" & X"55555555" & X"55555555" & X"55555555" & X"55555555" & X"55555555";
                    when 2 =>
                        return X"55555555" & X"55555555" & X"55555555" & X"55555555" & X"55555555" & X"55555555" & X"55555555" & X"55555555";
                    when 3 =>
                        return X"55555555" & X"55555555" & X"55555555" & X"55555555" & X"55555555" & X"55555555" & X"55555555" & X"55555555";
                    when others =>
                        -- synopsys translate_off
                        assert false
                            report "*** memif_ddrsdram_tuning_vector: unknown bank ( " & conv_string(bank) & ") for " &
                                memif_model_to_string(model)
                            severity failure;
                        -- synopsys translate_on
                        return "-";
                end case;

            when memif_model_adpxpi =>
                case bank is
                    when 0 =>
                        return X"55555555" & X"55555555" & X"55555555" & X"55555555" & X"55555555" & X"55595555" & X"55555555" & X"55555555";
                    when 1 =>
                        return X"55555555" & X"55aaa565" & X"55a99955" & X"65555555" & X"55666a55" & X"55565555" & X"55555555" & X"55555555";
                    when 2 =>
                        return X"6555a565" & X"66555555" & X"55555555" & X"55555555" & X"55555555" & X"55555555" & X"55555555" & X"55555555";
                    when 3 =>
                        return X"55555555" & X"55555555" & X"55555555" & X"55555555" & X"55555555" & X"55555555" & X"55555555" & X"55555555";
                    when others =>
                        -- synopsys translate_off
                        assert false
                            report "*** memif_ddrsdram_tuning_vector: unknown bank ( " & conv_string(bank) & ") for " &
                                memif_model_to_string(model)
                            severity failure;
                        -- synopsys translate_on
                        return "-";
                end case;

            when others =>
                -- synopsys translate_off
                assert false
                    report "*** memif_ddrsdram_tuning_vector: unknown model ( " & memif_model_to_string(model) & ")"
                    severity failure;
                -- synopsys translate_on
                return "-";
        end case;
    end;

    function memif_ddrsdram_supports_dm(
        constant model : in integer)
    return boolean is
    begin
        case model is
            when memif_model_adpdrc2 =>
                return false;

            when memif_model_adpwrc2 | memif_model_admxpl | memif_model_admxp | memif_model_adpxpi =>
                return true;

            when others =>
                -- synopsys translate_off
                assert false
                    report "*** memif_ddrsdram_supports_dm: unknown model ( " & memif_model_to_string(model) & ")"
                    severity failure;
                -- synopsys translate_on
                return false;
        end case;
    end;

    function memif_ddr2sdram_num_clock(
        constant model : in integer)
    return integer is
    begin
        case model is
            when memif_model_admxrc4fx | memif_model_admxrc5lx | memif_model_admxrc5t1 =>
                return 2;
            when others =>
                -- synopsys translate_off
                assert false
                    report "*** memif_ddr2sdram_num_clock: unknown model ( " & memif_model_to_string(model) & ")"
                    severity failure;
                -- synopsys translate_on
                return 0;
        end case;
    end;

    function memif_ddr2sdram_num_clock_en(
        constant model : in integer)
    return integer is
    begin
        case model is
            when memif_model_admxrc4fx | memif_model_admxrc5lx | memif_model_admxrc5t1 =>
                return 1;
            when others =>
                -- synopsys translate_off
                assert false
                    report "*** memif_ddr2sdram_num_clock_en: unknown model ( " & memif_model_to_string(model) & ")"
                    severity failure;
                -- synopsys translate_on
                return 0;
        end case;
    end;

    function memif_ddr2sdram_max_num_phys_bank(
        constant model : in integer)
    return integer is
    begin
        case model is
            when memif_model_admxrc4fx | memif_model_admxrc5lx | memif_model_admxrc5t1 =>
                return 1;
            when others =>
                -- synopsys translate_off
                assert false
                    report "*** memif_ddr2sdram_max_num_phys_bank: unknown model ( " & memif_model_to_string(model) & ")"
                    severity failure;
                -- synopsys translate_on
                return 0;
        end case;
    end;

    function memif_ddr2sdram_max_bank_width(
        constant model : in integer)
    return integer is
    begin
        case model is
            when memif_model_admxrc4fx | memif_model_admxrc5lx | memif_model_admxrc5t1 =>
                return 3;
            when others =>
                -- synopsys translate_off
                assert false
                    report "*** memif_ddrsdram_max_bank_width: unknown model ( " & memif_model_to_string(model) & ")"
                    severity failure;
                -- synopsys translate_on
                return 0;
        end case;
    end;

    function memif_ddr2sdram_max_rowcol_width(
        constant model : in integer)
    return integer is
    begin
        case model is
            when memif_model_admxrc4fx | memif_model_admxrc5t1 =>
                return 16;
            when memif_model_admxrc5lx =>
                return 15;
            when others =>
                -- synopsys translate_off
                assert false
                    report "*** memif_ddrsdram_max_rowcol_width: unknown model ( " & memif_model_to_string(model) & ")"
                    severity failure;
                -- synopsys translate_on
                return 0;
        end case;
    end;

end;
